
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifndef _STRUCTINFO_HPP_
#define _STRUCTINFO_HPP_

#include <string>
#include <vector>

#include "typeinfo.hpp"

namespace ps
{
	/**
	 * \brief Holds the declaration of a single member, used by StructInfo.
	 */
	class MemberInfo
	{
		public:
			std::string name;
			TypeInfo type;
			int offset;
			int size;
	};
	/**
	 * \brief Holds the declaration of a single member function, used by StructInfo.
	 */
	class FunctionInfo
	{
		public:
			std::string name;
			void *function;
			TypeInfo rettype;
			std::vector<TypeInfo> arguments;
	};
	/**
	 * \brief Class holding the definition of an external struct.
	 *
	 * The struct can have member variables and functions which then can be
	 * directly modified/called from scripts, as well as a constructor, a
	 * destructor and operators.
	 */
	class StructInfo
	{
		public:
			StructInfo();
			~StructInfo();
			
			void setSize(int size)
			{
				this->size = size;
			}
			
			/**
			 * \brief Sets the constructor of the struct.
			 *
			 * The constructor is called after the memory is reserved, the first
			 * argument is a pointer to the new struct.
			 */
			template <typename T> void setConstructor(void (*constructor)(T*))
			{
				this->constructor = (void*)constructor;
			}
			/**
			 * \brief Sets the destructor of the struct.
			 *
			 * The constructor is called right before the memory is freed, the
			 * first argument is a pointer to the struct.
			 */
			template <typename T> void setDestructor(void (*destructor)(T*))
			{
				this->destructor = (void*)destructor;
			}
			/**
			 * \brief Sets the assignment constructor of the struct.
			 *
			 * The first argument is a pointer to the struct, the second is the
			 * variable class (declared in module.h), the third is the type name
			 * (usable with structs).
			 *
			 * The forth argument holds the new value, it is either a pointer to
			 * a struct or can be casted directly to an integer or float,
			 * depending on the variable class like (int value = *((int*)&ptr)).
			 */
			template <typename T> void setOpEqual(void (*opequal)(T*, int, char*, void*))
			{
				this->opequal = (void*)opequal;
			}
			
			/**
			 * \brief Adds a member variable.
			 * \param name Name of the variable
			 * \param type Variable type
			 * \param offset Offset of the variable
			 * \param size Size of the variable
			 */
			void addMember(std::string name, TypeInfo type, int offset, int size);
			/**
			 * \brief Adds a member function.
			 *
			 * The first parameter of the function (which must not be included
			 * in "arguments") is a pointer to the struct.
			 * \param name Name of the function
			 * \param function Pointer to the function
			 * \param rettype Return type
			 * \param arguments Array with the type of the function arguments
			 */
			void addFunction(std::string name, void *function,
				TypeInfo rettype = TypeInfo(ETC_None),
				std::vector<TypeInfo> arguments = std::vector<TypeInfo>());
			
			int size;
			void *constructor;
			void *destructor;
			void *opequal;
			std::vector<MemberInfo> members;
			std::vector<FunctionInfo> functions;
	};
}

#endif

