
#ifndef _VM_HPP_
#define _VM_HPP_

#include "module.hpp"
#include "funcptr.hpp"
#include "typeinfo.hpp"
#include "structinfo.hpp"

#include <string>
#include <vector>

struct ps_vm_t;
struct struct_info_t;

namespace ps
{
	
	/**
	 * \brief Central class for a PlusScript virtual machine.
	 *
	 * The machine manages modules and external functions and types.
	 */
	class VM
	{
		public:
			VM();
			~VM();
			
			/**
			 * \brief Adds a module and generates bytecode from the source.
			 * \param filename Name of the file to read the module from
			 */
			Module *loadModule(const char *filename);
			/**
			 * \brief Loads previously generated bytecode. UNIMPLEMENTED.
			 * \param filename Name of the file to read the module from
			 */
			Module *loadBytecode(const char *filename);
			
			/**
			 * \brief Returns a pointer to a script function.
			 *
			 * The pointer does not point to the function itself but to a stub
			 * function which if needed compiles the function.
			 */
			FuncPtr getFunction(const char *name);
			
			/**
			 * \brief Registers an external function.
			 *
			 * The function can be used by scripts afterwards. Functions must
			 * be registered before compiling scripts which might need them.
			 */
			void registerFunction(std::string name, void *addr,
				TypeInfo rettype = TypeInfo(ETC_None),
				std::vector<TypeInfo> arguments = std::vector<TypeInfo>());
			
			/**
			 * \brief Registers an external struct.
			 *
			 * The struct can be used by scripts afterwards. Structs must
			 * be registered before compiling scripts which might need them.
			 */
			void registerStruct(std::string name, StructInfo info);
			
			void registerBaseFunctions(void);
			
			/**
			 * \brief Sets output for debug messages.
			 *
			 * If filename is 0 then stdout is used instead.
			 * \param filename File to write messages to
			 */
			static void setLoggingOutput(const char *filename);
			
			/**
			 * \brief Returns the vm struct.
			 */
			 ps_vm_t *getVM(void);
		private:
			ps_vm_t *vm;
			static void *logfile;
			
			std::vector<Module*> modules;
			std::vector<struct_info_t*> structinfo;
	};
}

#endif

