
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "plusscript.h"

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>

// Some output functions
void print(char *msg)
{
	printf("%s\n", msg);
}
void printint(int value)
{
	printf("%d\n", value);
}
void printfloat(float value)
{
	printf("%f\n", value);
}

// Some functions to test structs
typedef struct teststruct_t
{
	int a;
	float b;
} teststruct_t;

int testmemberfunc(teststruct_t *ts, int a, float b)
{
	printf("Old a: %d\n", ts->a);
	printf("Old b: %f\n", ts->b);
	ts->a = a;
	ts->b = b;
	printf("New a: %d\n", ts->a);
	printf("New b: %f\n", ts->b);
	return 42;
}
void testconstructor(teststruct_t *ts)
{
	printf("Constructor.\n");
	ts->a = 3;
	ts->b = 4;
}
void testdestructor(teststruct_t *ts)
{
	printf("Destructor (%d/%f).\n", ts->a, ts->b);
}
void testoperator(teststruct_t *ts, int type, char *type_name, void *data)
{
	printf("Operator '=' (%d).\n", type);
	if (type == VAR_TYPE_FLOAT)
	{
		ts->b = *((float*)&data);
	}
}

struct_info_t teststructinfo =
{
	"teststruct",
	sizeof(teststruct_t),
	(CONSTRUCTOR_PTR)testconstructor,
	(DESTRUCTOR_PTR)testdestructor,
	(OPERATOR_EQ_PTR)testoperator,
	2,
	0,
	1,
	0
};

// Test data
const char *tests[4] = {
"libplusscript/tests/if.ps",
"libplusscript/tests/float.ps",
"libplusscript/tests/struct.ps",
"libplusscript/tests/base.ps"
};
int teststatus[4];
int testcount = 4;

typedef int (*TESTFUNCTION)(void);

int main(int argc, char **argv)
{
	// Create structure info
	teststructinfo.members = (member_var_t*)malloc(2 * sizeof(member_var_t));
	teststructinfo.members[0].name = "a";
	teststructinfo.members[0].type = VAR_TYPE_INT;
	teststructinfo.members[0].type_name = 0;
	teststructinfo.members[0].offset = offsetof(teststruct_t, a);
	teststructinfo.members[0].size = sizeof(int);
	teststructinfo.members[1].name = "b";
	teststructinfo.members[1].type = VAR_TYPE_FLOAT;
	teststructinfo.members[1].type_name = 0;
	teststructinfo.members[1].offset = offsetof(teststruct_t, b);
	teststructinfo.members[1].size = sizeof(float);
	teststructinfo.functions = (ext_func_t*)malloc(sizeof(ext_func_t));
	teststructinfo.functions[0].name = "testfunc";
	teststructinfo.functions[0].function = (void*)testmemberfunc;
	teststructinfo.functions[0].argcount = 2;
	teststructinfo.functions[0].argtypes = (int*)malloc(3 * sizeof(int));
	teststructinfo.functions[0].argtypes[0] = VAR_TYPE_INT;
	teststructinfo.functions[0].argtypes[1] = VAR_TYPE_FLOAT;
	teststructinfo.functions[0].returntype = VAR_TYPE_INT;
	
	FILE *log = fopen("plusscript.log", "wt");
	set_logging_output(log);
	int i;
	for (i = 0; i < testcount; i++)
	{
		printf("======================================================\n");
		printf("Running test \"%s\".\n", tests[i]);
		printf("======================================================\n");
		// Create VM
		ps_vm_t *vm = create_vm();
		vm_register_base(vm);
		// Register functions
		int argtype = VAR_TYPE_USERPTR;
		vm_register_function(vm, (void*)print, "print", 1, &argtype, VAR_TYPE_UNKNOWN);
		argtype = VAR_TYPE_INT;
		vm_register_function(vm, (void*)printint, "printint", 1, &argtype, VAR_TYPE_UNKNOWN);
		argtype = VAR_TYPE_FLOAT;
		vm_register_function(vm, (void*)printfloat, "printfloat", 1, &argtype, VAR_TYPE_UNKNOWN);
		// Register test struct
		vm_register_struct(vm, &teststructinfo);
		
		module_t *testmod = parse_source(vm, tests[i]);
		//display_bytecode(testmod);
	
		create_stubs(vm, testmod);
		vm_add_module(vm, testmod);
	
		int (*testentry)(void);
		testentry = (TESTFUNCTION)vm_get_function_address(vm, "runtest");
		if (!testentry)
		{
			printf("Entry point not found.\n");
			teststatus[i] = 0;
		}
		else
		{
			teststatus[i] = testentry();
		}
		
		delete_module(vm, testmod);
		delete_vm(vm);
	}
	
	printf("\n");
	for (i = 0; i < testcount; i++)
	{
		printf("Test %d (%s): \t", i, tests[i]);
		if (teststatus[i])
		{
			printf("success.\n");
		}
		else
		{
			printf("failure.\n");
		}
	}
	
	if (log) fclose(log);
	
	return 0;
}

