
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifdef __amd64__
#include "amd64.h"
#include "bytecode.h"
#include "logger.h"

#include <stdio.h>

void amd64_set_local_var(module_t *module, int var, int reg)
{
	if (reg < REG_XMM0)
	{
		module->program[module->proglen++] = 0x48; // mov %reg, -var * 8(%rbp)
		module->program[module->proglen++] = 0x89;
		switch (reg)
		{
			case REG_RAX:
				module->program[module->proglen++] = 0x45;
				break;
			case REG_RDX:
				module->program[module->proglen++] = 0x55;
				break;
			case REG_RSI:
				module->program[module->proglen++] = 0x75;
				break;
			case REG_RDI:
				module->program[module->proglen++] = 0x7d;
				break;
			default:
				log_print(LOG_ERROR, "Error: amd64_set_local_var: Register not implemented.\n");
		}
		module->program[module->proglen++] = 0x100 - ((var + 1) * 8);
	}
	else
	{
		module->program[module->proglen++] = 0xf3; // movss %reg, -var * 8(%rbp)
		module->program[module->proglen++] = 0x0f;
		module->program[module->proglen++] = 0x11;
		switch (reg)
		{
			case REG_XMM0:
				module->program[module->proglen++] = 0x45;
				break;
			case REG_XMM1:
				module->program[module->proglen++] = 0x4d;
				break;
			case REG_XMM2:
				module->program[module->proglen++] = 0x55;
				break;
			case REG_XMM3:
				module->program[module->proglen++] = 0x5d;
				break;
			default:
				log_print(LOG_ERROR, "Error: amd64_set_local_var: Register not implemented.\n");
		}
		module->program[module->proglen++] = 0x100 - ((var + 1) * 8);
	}
}

void amd64_get_local_var(module_t *module, int var, int reg)
{
	module->program[module->proglen++] = 0x48; // mov -i * 8(%rbp), %reg
	module->program[module->proglen++] = 0x8b;
	switch (reg)
	{
		case REG_RAX:
			module->program[module->proglen++] = 0x45;
			break;
		case REG_RDX:
			module->program[module->proglen++] = 0x55;
			break;
		case REG_RSI:
			module->program[module->proglen++] = 0x75;
			break;
		case REG_RDI:
			module->program[module->proglen++] = 0x7d;
			break;
		default:
			log_print(LOG_ERROR, "Error: amd64_get_local_var: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x100 - ((var + 1) * 8);
}
void amd64_get_offset(module_t *module, int addrreg, int offset, int destreg, int size)
{
	if ((addrreg != REG_RAX) || (destreg != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_get_offset: Register not implemented.\n");
	}
	if (size < 8)
	{
		module->program[module->proglen++] = 0x48; // mov $0, %rcx
		module->program[module->proglen++] = 0x31;
		module->program[module->proglen++] = 0xc9;
	}
	switch (size)
	{
		case 1:
			module->program[module->proglen++] = 0x8a; // mov offset(%rax), %cl
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 2:
			module->program[module->proglen++] = 0x66; // mov offset(%rax), %cx
			module->program[module->proglen++] = 0x8b;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 4:
			module->program[module->proglen++] = 0x8b; // mov offset(%rax), %ecx
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 8:
			module->program[module->proglen++] = 0x48; // mov offset(%rax), %rcx
			module->program[module->proglen++] = 0x8b;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
	}
}
void amd64_set_offset(module_t *module, int addrreg, int offset, int valuereg, int size)
{
	if ((addrreg != REG_RAX) || (valuereg != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_get_offset: Register not implemented.\n");
	}
	switch (size)
	{
		case 1:
			module->program[module->proglen++] = 0x88; // mov %cl, offset(%rax)
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 2:
			module->program[module->proglen++] = 0x66; // mov %cx, offset(%rax)
			module->program[module->proglen++] = 0x89;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 4:
			module->program[module->proglen++] = 0x89; // mov %ecx, offset(%rax)
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 8:
			module->program[module->proglen++] = 0x48; // mov %rcx, offset(%rax)
			module->program[module->proglen++] = 0x89;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
	}

}
void amd64_call_function(module_t *module, void *function)
{
	module->program[module->proglen++] = 0x48; // mov $function, %rax
	module->program[module->proglen++] = 0xb8;
	*((void**)&module->program[module->proglen]) = function;
	module->proglen += sizeof(void*);
	module->program[module->proglen++] = 0xff; // call *%rax
	module->program[module->proglen++] = 0xd0;
}
void amd64_add(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %rax
		module->program[module->proglen++] = 0x59; // pop %rcx
		module->program[module->proglen++] = 0x48; // add %rcx, %rax
		module->program[module->proglen++] = 0x01;
		module->program[module->proglen++] = 0xc8;
		module->program[module->proglen++] = 0x50; // push %rax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		
		module->program[module->proglen++] = 0x48; // add $8, %rsp
		module->program[module->proglen++] = 0x83;
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = 0x08;
		
		module->program[module->proglen++] = 0xd8; // fadd (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: amd64_add: Type not implemented.\n");
	}
}
void amd64_sub(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %rax
		module->program[module->proglen++] = 0x59; // pop %rcx
		module->program[module->proglen++] = 0x48; // sub %rcx, %rax
		module->program[module->proglen++] = 0x29;
		module->program[module->proglen++] = 0xc8;
		module->program[module->proglen++] = 0x50; // push %rax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		
		module->program[module->proglen++] = 0x48; // add $8, %rsp
		module->program[module->proglen++] = 0x83;
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = 0x08;
		
		module->program[module->proglen++] = 0xd8; // fsub (%rsp)
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: amd64_sub: Type not implemented.\n");
	}
}
void amd64_mul(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %rax
		module->program[module->proglen++] = 0x59; // pop %rcx
		module->program[module->proglen++] = 0x48; // imul %rcx
		module->program[module->proglen++] = 0xf7;
		module->program[module->proglen++] = 0xe8;
		module->program[module->proglen++] = 0x50; // push %rax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		
		module->program[module->proglen++] = 0x48; // add $8, %rsp
		module->program[module->proglen++] = 0x83;
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = 0x08;
		
		module->program[module->proglen++] = 0xd8; // fmul (%rsp)
		module->program[module->proglen++] = 0x0c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: amd64_sub: Type not implemented.\n");
	}
}
void amd64_div(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %rax
		module->program[module->proglen++] = 0x59; // pop %rcx
		module->program[module->proglen++] = 0x48; // mov $0, %rdx
		module->program[module->proglen++] = 0x31;
		module->program[module->proglen++] = 0xd2;
		module->program[module->proglen++] = 0x48; // imul %rcx
		module->program[module->proglen++] = 0xf7;
		module->program[module->proglen++] = 0xf8;
		module->program[module->proglen++] = 0x50; // push %rax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		
		module->program[module->proglen++] = 0x48; // add $8, %rsp
		module->program[module->proglen++] = 0x83;
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = 0x08;
		
		module->program[module->proglen++] = 0xd8; // fdiv (%rsp)
		module->program[module->proglen++] = 0x34;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: amd64_sub: Type not implemented.\n");
	}
}
void amd64_and(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_and: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %rax
	module->program[module->proglen++] = 0x59; // pop %rcx
	module->program[module->proglen++] = 0x48; // and %rcx, %rax
	module->program[module->proglen++] = 0x21;
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %rax
}
void amd64_or(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_or: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %rax
	module->program[module->proglen++] = 0x59; // pop %rcx
	module->program[module->proglen++] = 0x48; // and %rcx, %rax
	module->program[module->proglen++] = 0x09;
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %rax
}
void amd64_xor(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_xor: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %rax
	module->program[module->proglen++] = 0x59; // pop %rcx
	module->program[module->proglen++] = 0x48; // and %rcx, %rax
	module->program[module->proglen++] = 0x31;
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %rax
}
void amd64_compare(module_t *module, int reg1, int reg2, int type, int vartype)
{
	if ((reg1 != REG_RAX) || (reg2 != REG_RCX))
	{
		log_print(LOG_ERROR, "Error: amd64_compare: Register not implemented.\n");
	}
	
	if (vartype == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %rax
		module->program[module->proglen++] = 0x59; // pop %rcx
		module->program[module->proglen++] = 0x6a; // push $0
		module->program[module->proglen++] = 0x00;
		module->program[module->proglen++] = 0x48; // cmp %rax, %rcx
		module->program[module->proglen++] = 0x39;
		module->program[module->proglen++] = 0xc1;
	}
	else if (vartype == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xf3; // movss (%rsp),%xmm0
		module->program[module->proglen++] = 0x0f;
		module->program[module->proglen++] = 0x10;
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x48; // add $8, %rsp
		module->program[module->proglen++] = 0x83;
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = 0x08;
		module->program[module->proglen++] = 0x0f; // ucomiss (%rsp), %xmm0
		module->program[module->proglen++] = 0x2e;
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x48; // mov $0, (%rsp)
		module->program[module->proglen++] = 0xc7;
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x00;
		module->program[module->proglen++] = 0x00;
		module->program[module->proglen++] = 0x00;
		module->program[module->proglen++] = 0x00;
	}
	else
	{
		log_print(LOG_ERROR, "Error: amd64_compare: Type not implemented (%d).\n", vartype);
	}
	
	if (type == CMP_SMALLER)
	{
		module->program[module->proglen++] = 0x7d; // jge
	}
	else if (type == CMP_GREATER)
	{
		module->program[module->proglen++] = 0x7e; // jle
	}
	else if (type == (CMP_SMALLER | CMP_EQUAL))
	{
		module->program[module->proglen++] = 0x7f; // jg
	}
	else if (type == (CMP_GREATER | CMP_EQUAL))
	{
		module->program[module->proglen++] = 0x7c; // jl
	}
	else if (type == (CMP_GREATER | CMP_SMALLER))
	{
		module->program[module->proglen++] = 0x74; // je
	}
	else if (type == CMP_EQUAL)
	{
		module->program[module->proglen++] = 0x75; // jne
	}
	module->program[module->proglen++] = 0x04;
	
	module->program[module->proglen++] = 0xc6; // mov $1, (%esp)
	module->program[module->proglen++] = 0x04;
	module->program[module->proglen++] = 0x24;
	module->program[module->proglen++] = 0x01;
}

void amd64_convert(module_t *module, int srctype, int desttype)
{
	log_print(LOG_DEBUG, "Converting %d to %d.\n", srctype, desttype);
	if ((srctype == VAR_TYPE_FLOAT) && (desttype == VAR_TYPE_INT))
	{
		log_print(LOG_DEBUG, "Converting float.\n");
		module->program[module->proglen++] = 0x50; // push %rax
		module->program[module->proglen++] = 0xd9; // flds (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xdb; // fistp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x58; // pop %rax
	}
	else if ((srctype == VAR_TYPE_INT) && (desttype == VAR_TYPE_FLOAT))
	{
		module->program[module->proglen++] = 0x50; // push %rax
		module->program[module->proglen++] = 0xdf; // fild (%rsp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%rsp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x58; // pop %rax
	}
	else
	{
		log_print(LOG_WARNING, "Warning: amd64_convert: Cannot convert %d to %d.\n", srctype, desttype);
	}
}

void amd64_set_int_argument(module_t *module, int idx)
{
	if (idx > 5)
	{
		log_print(LOG_ERROR, "Error: amd64_int_argument: Too many arguments.\n");
		module->program[module->proglen++] = 0x58; // pop %rax
	}
	else
	{
		switch (idx)
		{
			case 0:
				module->program[module->proglen++] = 0x5f; // pop %rdi
				break;
			case 1:
				module->program[module->proglen++] = 0x5e; // pop %rsi
				break;
			case 2:
				module->program[module->proglen++] = 0x5a; // pop %rdx
				break;
			case 3:
				module->program[module->proglen++] = 0x59; // pop %rcx
				break;
			case 4:
				module->program[module->proglen++] = 0x41; // pop %r8
				module->program[module->proglen++] = 0x58;
				break;
			case 5:
				module->program[module->proglen++] = 0x41; // pop %r9
				module->program[module->proglen++] = 0x59;
				break;
		}
	}
}
void amd64_set_float_argument(module_t *module, int idx)
{
	if (idx > 3)
	{
		log_print(LOG_ERROR, "Error: amd64_float_argument: Too many arguments.\n");
	}
	else
	{
		switch (idx)
		{
			case 0:
				module->program[module->proglen++] = 0xf3; // mov (%rsp), %xmm0
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x10;
				module->program[module->proglen++] = 0x04;
				module->program[module->proglen++] = 0x24;
				break;
			case 1:
				module->program[module->proglen++] = 0xf3; // mov (%rsp), %xmm1
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x10;
				module->program[module->proglen++] = 0x0c;
				module->program[module->proglen++] = 0x24;
				break;
			case 2:
				module->program[module->proglen++] = 0xf3; // mov (%rsp), %xmm2
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x10;
				module->program[module->proglen++] = 0x14;
				module->program[module->proglen++] = 0x24;
				break;
			case 3:
				module->program[module->proglen++] = 0xf3; // mov (%rsp), %xmm3
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x10;
				module->program[module->proglen++] = 0x1c;
				module->program[module->proglen++] = 0x24;
				break;
		}
	}
	module->program[module->proglen++] = 0x58; // pop %rax
}
void amd64_get_int_argument(module_t *module, int var, int idx)
{
	switch (idx)
	{
		case 0:
			amd64_set_local_var(module, var, REG_RDI);
			break;
		case 1:
			amd64_set_local_var(module, var, REG_RSI);
			break;
		case 2:
			amd64_set_local_var(module, var, REG_RDX);
			break;
	}
}
void amd64_get_float_argument(module_t *module, int var, int idx)
{
	switch (idx)
	{
		case 0:
			amd64_set_local_var(module, var, REG_XMM0);
			break;
		case 1:
			amd64_set_local_var(module, var, REG_XMM1);
			break;
		case 2:
			amd64_set_local_var(module, var, REG_XMM2);
			break;
		case 3:
			amd64_set_local_var(module, var, REG_XMM3);
			break;
	}
}

void amd64_call_operator(module_t *module, int varidx, void *function, int type, char *name, int local)
{
	module->program[module->proglen++] = 0x48; //mov %rax, %rcx
	module->program[module->proglen++] = 0x89;
	module->program[module->proglen++] = 0xc1;
	module->program[module->proglen++] = 0x48; //mov $name, %rcx
	module->program[module->proglen++] = 0xba;
	*((char**)&module->program[module->proglen]) = name;
	module->proglen += sizeof(char*);
	module->program[module->proglen++] = 0x48; //mov $type, %rsi
	module->program[module->proglen++] = 0xc7;
	module->program[module->proglen++] = 0xc6;
	*((int*)&module->program[module->proglen]) = type;
	module->proglen += sizeof(int);
	if (local)
	{
		amd64_get_local_var(module, varidx, REG_RAX);
		module->program[module->proglen++] = 0x48; //mov %rax, %rdi
		module->program[module->proglen++] = 0x89;
		module->program[module->proglen++] = 0xc7;
	}
	else
	{
		void *addr = module->data + varidx * 4;
		module->program[module->proglen++] = 0xa1; // mov var, %rax
		*((void**)&module->program[module->proglen]) = addr;
		module->proglen += sizeof(void*);
		module->program[module->proglen++] = 0x48; //mov %rax, %rdi
		module->program[module->proglen++] = 0x89;
		module->program[module->proglen++] = 0xc7;
	}
	amd64_call_function(module, function);
}

#endif

