
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "compiler.h"
#include "bytecode.h"
#include "logger.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(__x86_64__) && !defined(__amd64__)
#define __amd64__
#endif

#include "amd64.h"
#include "x86.h"

#ifdef __unix__
#include <sys/mman.h>
#endif

// TODO: Variable code size
#define MODULE_CODE_SIZE 0x1000

void create_stubs(ps_vm_t *vm, module_t *module)
{
	int i;
	for (i = 0; i < module->funccount; i++)
	{
		// Allocate executable memory
		#ifdef __unix__
		module->functions[i].call_stub = (char*)mmap(0, 0x1000,
			PROT_READ | PROT_WRITE | PROT_EXEC,
			MAP_PRIVATE | MAP_ANON,
			-1, 0);
		#else
		module->functions[i].call_stub = (char*)malloc(512);
		#endif
		
		#if defined(__amd64__)
		// Write stub
		char *stub = module->functions[i].call_stub;
		int stubsize = 0;
		stub[stubsize++] = 0x57; // push %rdi
		stub[stubsize++] = 0x56; // push %rsi
		stub[stubsize++] = 0x52; // push %rdx
		stub[stubsize++] = 0x50; // push %rax
		stub[stubsize++] = 0x48; // mov $vm,%rdi
		stub[stubsize++] = 0xbf;
		*((ps_vm_t**)&stub[stubsize]) = vm;
		stubsize += sizeof(ps_vm_t*);
		stub[stubsize++] = 0x48; // mov $module,%rsi
		stub[stubsize++] = 0xbe;
		*((module_t**)&stub[stubsize]) = module;
		stubsize += sizeof(module_t*);
		stub[stubsize++] = 0x48; // mov $i,%rdx
		stub[stubsize++] = 0xc7;
		stub[stubsize++] = 0xc2;
		*((int*)&stub[stubsize]) = i;
		stubsize += sizeof(int);
		
		stub[stubsize++] = 0x48;  // mov $compile_function, %rax
		stub[stubsize++] = 0xb8;
		*((void**)&stub[stubsize]) = (void*)compile_function;
		stubsize += sizeof(void*);
		
		stub[stubsize++] = 0xff; // call *%rax
		stub[stubsize++] = 0xd0;
		
		stub[stubsize++] = 0x58; // pop %rax
		stub[stubsize++] = 0x5a; // pop %rdx
		stub[stubsize++] = 0x5e; // pop %rsi
		stub[stubsize++] = 0x5f; // pop %rdi
		// Jump back to the (now modified) beginning
		stub[stubsize++] = 0xeb;
		stub[stubsize++] = 0xcf;
		
		#elif defined(__i386__)
		char *stub = module->functions[i].call_stub;
		int stubsize = 0;
		stub[stubsize++] = 0x50; // push %eax
		stub[stubsize++] = 0x68; // push $i
		*((int*)&stub[stubsize]) = i;
		stubsize += sizeof(int);
		stub[stubsize++] = 0x68; // push $module
		*((module_t**)&stub[stubsize]) = module;
		stubsize += sizeof(module_t*);
		stub[stubsize++] = 0x68; // push $vm
		*((ps_vm_t**)&stub[stubsize]) = vm;
		stubsize += sizeof(ps_vm_t*);
		stub[stubsize++] = 0xb8; // mov $compile_function, %eax
		*((void**)&stub[stubsize]) = (void*)compile_function;
		stubsize += sizeof(void*);
		stub[stubsize++] = 0xff; // call *%eax
		stub[stubsize++] = 0xd0;
		stub[stubsize++] = 0x83; // add $0xc, %esp
		stub[stubsize++] = 0xc4;
		stub[stubsize++] = 0x0c;
		stub[stubsize++] = 0x58; // pop %eax
		// Jump back to the (now modified) beginning
		stub[stubsize++] = 0xeb;
		stub[stubsize++] = 0xe3;
		
		#else
		#error Architecture not supported!
		#endif
	}
}
void compile_function(ps_vm_t *vm, module_t *module, int index)
{
	mod_func_t *func = &module->functions[index];
	log_print(LOG_INFORMATION, "Compiling function \"%s\"\n", module->strings[func->name]);
	
	func->progoffset = module->proglen;
	
	if (!module->program)
	{
		#ifdef __unix__
		module->program = (char*)mmap(0, MODULE_CODE_SIZE,
			PROT_READ | PROT_WRITE | PROT_EXEC,
			MAP_PRIVATE | MAP_ANON,
			-1, 0);
		log_print(LOG_DEBUG, "Program: %p\n", module->program);
		#else
		module->program = (char*)malloc(MODULE_CODE_SIZE);
		#endif
	}
	
	// Enter function
	#ifdef __amd64__
	module->program[module->proglen++] = 0x55; // push %rbp
	module->program[module->proglen++] = 0x48; // mov %rsp, %rbp
	module->program[module->proglen++] = 0x89;
	module->program[module->proglen++] = 0xe5;
	#endif
	#ifdef __i386__
	module->program[module->proglen++] = 0x55; // push %ebp
	module->program[module->proglen++] = 0x89; // mov %esp, %ebp
	module->program[module->proglen++] = 0xe5;
	#endif
	// Get space for local variables
	#ifdef __amd64__
	module->program[module->proglen++] = 0x48; // sub $func->varcount, %rsp
	module->program[module->proglen++] = 0x81;
	module->program[module->proglen++] = 0xec;
	if (func->varcount % 2)
	{
		*((int*)&module->program[module->proglen]) = func->varcount * 8;
	}
	else
	{
		*((int*)&module->program[module->proglen]) = func->varcount * 8 + 8;
	}
	module->proglen += sizeof(int);
	#endif
	#ifdef __i386__
	x86_adjust_esp(module, -func->varcount * 4);
	#endif
	// Get function arguments
	int i;
	int varidx = -1;
	log_print(LOG_DEBUG, "%d arguments.\n", func->argcount);
	#ifdef __amd64__
	int intargcount = 0;
	int floatargcount = 0;
	for (i = 0; i < func->argcount; i++)
	{
		varidx = get_variable_index(func, func->argnames[i]);
		if (varidx != -1)
		{
			if (func->vars[varidx].type == VAR_TYPE_FLOAT)
			{
				amd64_get_float_argument(module, varidx, floatargcount);
				floatargcount++;
			}
			else
			{
				amd64_get_int_argument(module, varidx, intargcount);
				intargcount++;
			}
		}
		varidx = -1;
	}
	#endif
	#ifdef __i386__
	for (i = 0; i < func->argcount; i++)
	{
		varidx = get_variable_index(func, func->argnames[i]);
		if (varidx != -1)
		{
			module->program[module->proglen++] = 0x8b; // mov i(%ebp), %eax
			module->program[module->proglen++] = 0x45;
			module->program[module->proglen++] = (i + 2) * 4;
			x86_set_local_var(module, varidx, REG_EAX);
		}
	}
	#endif
	// Allocate space and call constructors
	for (i = 0; i < func->varcount; i++)
	{
		if (func->vars[i].type == VAR_TYPE_STRUCT)
		{
			struct_info_t *info = vm_get_struct_info(vm, module->strings[func->vars[i].type_name]);
			if (info)
			{
				// Call malloc
				#ifdef __amd64__
				module->program[module->proglen++] = 0x48; // mov $size, %rdi
				module->program[module->proglen++] = 0xc7;
				module->program[module->proglen++] = 0xc7;
				*((int*)&module->program[module->proglen]) = info->size;
				module->proglen += sizeof(int);
				amd64_call_function(module, (void*)malloc);
				amd64_set_local_var(module, i, REG_RAX);
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x68; // push $size
				*((int*)&module->program[module->proglen]) = info->size;
				module->proglen += sizeof(int);
				x86_call_function(module, (void*)malloc);
				x86_adjust_esp(module, 4);
				x86_set_local_var(module, i, REG_EAX);
				#endif
				// Call constructor
				if (info->constructor)
				{
					#ifdef __amd64__
					module->program[module->proglen++] = 0x48; // mov %rax, %rdi
					module->program[module->proglen++] = 0x89;
					module->program[module->proglen++] = 0xc7;
					amd64_call_function(module, (void*)info->constructor);
					#endif
					#ifdef __i386__
					module->program[module->proglen++] = 0x50; // push %eax
					x86_call_function(module, (void*)info->constructor);
					x86_adjust_esp(module, 4);
					#endif
				}
			}
		}
	}
	
	// Unresolved jump labels
	char **jmpaddr = 0;
	int *jmplabel = 0;
	int unrescount = 0;
	// Labels
	char **labeladdr = 0;
	int *labelidx = 0;
	int labelcount = 0;
	
	int *vartype = 0;
	int vartypecount = 0;
	char *currenttype = 0;
	
	int pos = func->bcoffset;
	while ((pos < module->codelen) && (module->bytecode[pos] != CMD_END))
	{
		log_print(LOG_DEBUG, "vartypecount: %d\n", vartypecount);
		switch (module->bytecode[pos])
		{
			case CMD_IF:
			{
				int label = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: IF (%d)\n", pos, label);
				
				// Test variable
				#ifdef __amd64__
				module->program[module->proglen++] = 0x58; // pop %rax
				module->program[module->proglen++] = 0x48; // cmp $0, %rax
				module->program[module->proglen++] = 0x83;
				module->program[module->proglen++] = 0xf8;
				module->program[module->proglen++] = 0x00;
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x58; // pop %eax
				module->program[module->proglen++] = 0x83; // cmp $0, %eax
				module->program[module->proglen++] = 0xf8;
				module->program[module->proglen++] = 0x00;
				#endif
				// Jump
				#ifdef __amd64__
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x84;
				*((int*)&module->program[module->proglen]) = 0;
				module->proglen += 4;
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x0f;
				module->program[module->proglen++] = 0x84;
				*((int*)&module->program[module->proglen]) = 0;
				module->proglen += 4;
				#endif
				vartypecount--;
				
				// Search for label
				int resolved = 0;
				for (i = 0; i < labelcount; i++)
				{
					if (labelidx[i] == label)
					{
						#ifdef __amd64__
						long long addrdiff = labeladdr[i] - &module->program[module->proglen - 4];
						*((int*)&module->program[module->proglen - 4]) = addrdiff;
						#endif
						#ifdef __i386__
						int addrdiff = labeladdr[i] - &module->program[module->proglen - 4];
						*((int*)&module->program[module->proglen - 4]) = addrdiff;
						#endif
						resolved = 1;
						break;
					}
				}
				if (!resolved)
				{
					// Not found, adding it to the unresolved list
					jmpaddr = (char**)realloc(jmpaddr, sizeof(char*) * (unrescount + 1));
					jmpaddr[unrescount] = &module->program[module->proglen - 4];
					jmplabel = (int*)realloc(jmplabel, sizeof(int) * (unrescount + 1));
					jmplabel[unrescount] = label;
					unrescount++;
				}
				
				pos += 5;
				break;
			}
			case CMD_CMP:
			{
				log_print(LOG_DEBUG, "%04d: CMP (", pos);
				char type = module->bytecode[pos + 1];
				if (type & CMP_SMALLER) log_print(LOG_DEBUG, "<");
				if (type & CMP_GREATER) log_print(LOG_DEBUG, ">");
				if (type & CMP_EQUAL) log_print(LOG_DEBUG, "=");
				if (type == CMP_EQUAL) log_print(LOG_DEBUG, "=");
				log_print(LOG_DEBUG, ")\n");
				
				if (vartype[vartypecount - 1] != vartype[vartypecount - 2])
				{
					module->program[module->proglen++] = 0x58; // pop %eax
					#ifdef __amd64__
					amd64_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					#ifdef __i386__
					x86_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					module->program[module->proglen++] = 0x50; // push %eax
				}
				
				#ifdef __amd64__
				amd64_compare(module, REG_RAX, REG_RCX, type, vartype[vartypecount - 2]);
				#endif
				#ifdef __i386__
				x86_compare(module, REG_EAX, REG_ECX, type, vartype[vartypecount - 2]);
				#endif
				
				vartypecount--;
				vartype[vartypecount - 1] = VAR_TYPE_INT;
				
				pos += 2;
				break;
			}
			case CMD_LABEL:
			{
				int label = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: LABEL (%d)\n", pos, label);
				
				// Resolve unresoved jumps
				for (i = 0; i < unrescount; i++)
				{
					if (jmplabel[i] == label)
					{
						// TODO: Remove from list?
						#ifdef __amd64__
						long long addrdiff = &module->program[module->proglen] - (jmpaddr[i] + 4);
						*((int*)jmpaddr[i]) = addrdiff;
						#endif
						#ifdef __i386__
						int addrdiff = &module->program[module->proglen] - (jmpaddr[i] + 4);
						*((int*)jmpaddr[i]) = addrdiff;
						#endif
					}
				}
				// Add label to list
				labeladdr = (char**)realloc(labeladdr, sizeof(char*) * (labelcount + 1));
				labeladdr[labelcount] = &module->program[module->proglen];
				labelidx = (int*)realloc(labelidx, sizeof(int) * (labelcount + 1));
				labelidx[labelcount] = label;
				labelcount++;
				
				pos += 5;
				break;
			}
			case CMD_JMP:
			{
				int label = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: JMP (label: %d)\n", pos, label);
				
				#ifdef __amd64__
				module->program[module->proglen++] = 0xe9;
				*((int*)&module->program[module->proglen]) = 0;
				module->proglen += 4;
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0xe9;
				*((int*)&module->program[module->proglen]) = 0;
				module->proglen += 4;
				#endif
				// Search for label
				int resolved = 0;
				for (i = 0; i < labelcount; i++)
				{
					if (labelidx[i] == label)
					{
						#ifdef __amd64__
						long long addrdiff = labeladdr[i] - &module->program[module->proglen];
						*((int*)&module->program[module->proglen - 4]) = addrdiff;
						#endif
						#ifdef __i386__
						int addrdiff = labeladdr[i] - &module->program[module->proglen];
						*((int*)&module->program[module->proglen - 4]) = addrdiff;
						#endif
						resolved = 1;
						break;
					}
				}
				if (!resolved)
				{
					// Not found, adding it to the unresolved list
					jmpaddr = (char**)realloc(jmpaddr, sizeof(char*) * (unrescount + 1));
					jmpaddr[unrescount] = &module->program[module->proglen - 4];
					jmplabel = (int*)realloc(jmplabel, sizeof(int) * (unrescount + 1));
					jmplabel[unrescount] = label;
					unrescount++;
				}
				
				pos += 5;
				break;
			}
			case CMD_SET:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: SET (%s)\n", pos, module->strings[var]);
				
				module->program[module->proglen++] = 0x58; // pop %eax
				
				// Get local variable
				int varidx = get_variable_index(func, var);
				if (varidx != -1)
				{
					if (func->vars[varidx].type != VAR_TYPE_STRUCT)
					{
						if (vartype[vartypecount - 1] != func->vars[varidx].type)
						{
							// Convert type
							#ifdef __amd64__
							amd64_convert(module, vartype[vartypecount - 1], func->vars[varidx].type);
							#endif
							#ifdef __i386__
							x86_convert(module, vartype[vartypecount - 1], func->vars[varidx].type);
							#endif
						}
						#ifdef __amd64__
						amd64_set_local_var(module, varidx, REG_RAX);
						#endif
						#ifdef __i386__
						x86_set_local_var(module, varidx, REG_EAX);
						#endif
					}
					else
					{
						char *structname = module->strings[func->vars[varidx].type_name];
						struct_info_t *info = vm_get_struct_info(vm, structname);
						if (!info)
						{
							log_print(LOG_ERROR, "Error: Structure info for \"%s\" not found.\n", structname);
						}
						else
						{
							if (info->op_equal)
							{
								// Call operator
								#ifdef __i386__
								x86_call_operator(module, varidx, (void*)info->op_equal, vartype[vartypecount - 1], structname, 1);
								#endif
								#ifdef __amd64__
								amd64_call_operator(module, varidx, (void*)info->op_equal, vartype[vartypecount - 1], structname, 1);
								#endif
							}
							else
							{
								log_print(LOG_ERROR, "Error: Structure \"%s\" does not have any operator '='.\n", structname);
							}
						}
					}
					vartypecount--;
				}
				else
				{
					// Get global variable
					for (i = 0; i < module->varcount; i++)
					{
						if (module->vars[i].name == var)
						{
							varidx = i;
							break;
						}
					}
					if (varidx != -1)
					{
						if (module->vars[varidx].type != VAR_TYPE_STRUCT)
						{
							if (vartype[vartypecount - 1] != module->vars[varidx].type)
							{
								// Convert type
								#ifdef __amd64__
								amd64_convert(module, vartype[vartypecount - 1], module->vars[varidx].type);
								#endif
								#ifdef __i386__
								x86_convert(module, vartype[vartypecount - 1], module->vars[varidx].type);
								#endif
							}
							#ifdef __amd64__
							void *addr = module->data + varidx * 8;
							module->program[module->proglen++] = 0x48; // mov %rax, addr
							module->program[module->proglen++] = 0xa3;
							*((void**)&module->program[module->proglen]) = addr;
							module->proglen += sizeof(void*);
							#endif
							#ifdef __i386__
							void *addr = module->data + varidx * 4;
							module->program[module->proglen++] = 0xa3; // mov %eax, addr
							*((void**)&module->program[module->proglen]) = addr;
							module->proglen += sizeof(void*);
							#endif
						}
						else
						{
							char *structname = module->strings[module->vars[varidx].type_name];
							struct_info_t *info = vm_get_struct_info(vm, structname);
							if (!info)
							{
								log_print(LOG_ERROR, "Error: Structure info for \"%s\" not found.\n", structname);
							}
							else
							{
								if (info->op_equal)
								{
									// Call operator
									#ifdef __i386__
									x86_call_operator(module, varidx, (void*)info->op_equal, vartype[vartypecount - 1], structname, 0);
									#endif
									#ifdef __amd64__
									amd64_call_operator(module, varidx, (void*)info->op_equal, vartype[vartypecount - 1], structname, 0);
									#endif
								}
								else
								{
									log_print(LOG_ERROR, "Error: Structure \"%s\" does not have any operator '='.\n", structname);
								}
							}
						}
						vartypecount--;
					}
					else
					{
						log_print(LOG_ERROR, "Error: Variable not found.\n");
					}
				}
				pos += 5;
				break;
			}
			case CMD_SETMEMBER:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: SETMEMBER (%s, %s)\n", pos, currenttype, module->strings[var]);
				
				struct_info_t *info = vm_get_struct_info(vm, currenttype);
				if (!info)
				{
					log_print(LOG_ERROR, "Error: Structure info for \"%s\" not found.\n", currenttype);
				}
				else
				{
					int memberidx = -1;
					for (i = 0; i < info->membercount; i++)
					{
						if (!strcmp(module->strings[var], info->members[i].name))
						{
							memberidx = i;
							break;
						}
					}
					if (memberidx == -1)
					{
						log_print(LOG_ERROR, "Error: Member not found.\n");
					}
					else
					{
						if (vartype[vartypecount - 2] != info->members[i].type)
						{
							module->program[module->proglen++] = 0x59; // pop %ecx
							module->program[module->proglen++] = 0x58; // pop %eax
							// Convert type
							#ifdef __amd64__
							amd64_convert(module, vartype[vartypecount - 2], info->members[i].type);
							#endif
							#ifdef __i386__
							x86_convert(module, vartype[vartypecount - 2], info->members[i].type);
							#endif
							module->program[module->proglen++] = 0x50; // push %eax
							module->program[module->proglen++] = 0x51; // push %ecx
						}
						module->program[module->proglen++] = 0x58; // pop %eax
						module->program[module->proglen++] = 0x59; // pop %ecx
						#ifdef __amd64__
						amd64_set_offset(module, REG_RAX, info->members[i].offset, REG_RCX, info->members[i].size);
						#endif
						#ifdef __i386__
						x86_set_offset(module, REG_EAX, info->members[i].offset, REG_ECX, info->members[i].size);
						#endif
					}
				}
				vartypecount-= 2;
				
				pos += 5;
				break;
			}
			case CMD_GET:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: GET (%s)\n", pos, module->strings[var]);
				
				int varidx = -1;
				// Get local variable
				for (i = 0; i < func->varcount; i++)
				{
					if (func->vars[i].name == var)
					{
						varidx = i;
						break;
					}
				}
				if (varidx != -1)
				{
					#ifdef __amd64__
					amd64_get_local_var(module, varidx, REG_RAX);
					#endif
					#ifdef __i386__
					x86_get_local_var(module, varidx, REG_EAX);
					#endif
					
					currenttype = module->strings[func->vars[varidx].type_name];
					vartypecount++;
					vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
					vartype[vartypecount - 1] = func->vars[varidx].type;
				}
				else
				{
					// Get global variable
					for (i = 0; i < module->varcount; i++)
					{
						if (module->vars[i].name == var)
						{
							varidx = i;
							break;
						}
					}
					if (varidx != -1)
					{
						#ifdef __amd64__
						void *addr = module->data + varidx * 8;
						module->program[module->proglen++] = 0x48; // mov addr, %rax
						module->program[module->proglen++] = 0xa1;
						*((void**)&module->program[module->proglen]) = addr;
						module->proglen += sizeof(void*);
						#endif
						#ifdef __i386__
						void *addr = module->data + varidx * 4;
						module->program[module->proglen++] = 0xa1; // mov addr, %rax
						*((void**)&module->program[module->proglen]) = addr;
						module->proglen += sizeof(void*);
						#endif
						
						currenttype = module->strings[module->vars[varidx].type_name];
						vartypecount++;
						vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
						vartype[vartypecount - 1] = module->vars[varidx].type;
					}
					else
					{
						log_print(LOG_ERROR, "Error: Variable not found.\n");
					}
				}
				
				#ifdef __amd64__
				module->program[module->proglen++] = 0x50; // push %rax
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x50; // push %eax
				#endif
				
				pos += 5;
				break;
			}
			case CMD_GETMEMBER:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: GETMEMBER (%s, %s)\n", pos, currenttype, module->strings[var]);
				
				struct_info_t *info = vm_get_struct_info(vm, currenttype);
				if (!info)
				{
					log_print(LOG_ERROR, "Error: Structure info for \"%s\" not found.\n", currenttype);
				}
				else
				{
					int memberidx = -1;
					for (i = 0; i < info->membercount; i++)
					{
						if (!strcmp(module->strings[var], info->members[i].name))
						{
							memberidx = i;
							break;
						}
					}
					if (memberidx == -1)
					{
						log_print(LOG_ERROR, "Error: Member not found.\n");
					}
					else
					{
						#ifdef __amd64__
						module->program[module->proglen++] = 0x58; // pop %rax
						amd64_get_offset(module, REG_RAX, info->members[memberidx].offset, REG_RCX, info->members[memberidx].size);
						module->program[module->proglen++] = 0x51; // push %rcx
						#endif
						#ifdef __i386__
						module->program[module->proglen++] = 0x58; // pop %eax
						x86_get_offset(module, REG_EAX, info->members[memberidx].offset, REG_ECX, info->members[memberidx].size);
						module->program[module->proglen++] = 0x51; // push %ecx
						#endif
					}
					vartype[vartypecount - 1] = info->members[memberidx].type;
				}
				
				pos += 5;
				break;
			}
			case CMD_GETC:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: GETC (%d)\n", pos, var);
				
				#ifdef __amd64__
				module->program[module->proglen++] = 0x48; // mov $var,%rax
				module->program[module->proglen++] = 0xb8;
				*((int*)&module->program[module->proglen]) = var;
				module->proglen += sizeof(int);
				*((int*)&module->program[module->proglen + 4]) = 0;
				module->proglen += sizeof(int);
				module->program[module->proglen++] = 0x50; // push %rax
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x68; // push $var
				*((int*)&module->program[module->proglen]) = var;
				module->proglen += sizeof(int);
				#endif
				
				vartypecount++;
				vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
				vartype[vartypecount - 1] = VAR_TYPE_INT;
				
				pos += 5;
				break;
			}
			case CMD_GETF:
			{
				float var = *((float*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: GETF\n", pos);
				
				#ifdef __amd64__
				module->program[module->proglen++] = 0x48; // mov $var,%rax
				module->program[module->proglen++] = 0xb8;
				*((float*)&module->program[module->proglen]) = var;
				module->proglen += sizeof(int);
				*((int*)&module->program[module->proglen + 4]) = 0;
				module->proglen += sizeof(int);
				module->program[module->proglen++] = 0x50; // push %rax
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x68; // push $var
				*((float*)&module->program[module->proglen]) = var;
				module->proglen += sizeof(float);
				#endif
				
				vartypecount++;
				vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
				vartype[vartypecount - 1] = VAR_TYPE_FLOAT;
				
				pos += 5;
				break;
			}
			case CMD_GETS:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: GETS (%d, \"%s\")\n", pos, var, module->strings[var]);
				
				#ifdef __amd64__
				module->program[module->proglen++] = 0x48; // mov $var,%rax
				module->program[module->proglen++] = 0xb8;
				*((char**)&module->program[module->proglen]) = module->strings[var];
				module->proglen += sizeof(char*);
				module->program[module->proglen++] = 0x50; // push %rax
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0x68; // push $var
				*((char**)&module->program[module->proglen]) = module->strings[var];
				module->proglen += sizeof(char*);
				#endif
				
				vartypecount++;
				vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
				vartype[vartypecount - 1] = VAR_TYPE_USERPTR;
				
				pos += 5;
				break;
			}
			case CMD_ADD:
			{
				log_print(LOG_DEBUG, "%04d: ADD\n", pos);
				if (vartype[vartypecount - 1] != vartype[vartypecount - 2])
				{
					module->program[module->proglen++] = 0x58; // pop %eax
					#ifdef __amd64__
					amd64_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					#ifdef __i386__
					x86_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					module->program[module->proglen++] = 0x50; // push %eax
				}
				#ifdef __amd64__
				amd64_add(module, REG_RAX, REG_RCX, vartype[vartypecount - 2]);
				#endif
				#ifdef __i386__
				x86_add(module, REG_EAX, REG_ECX, vartype[vartypecount - 2]);
				#endif
				
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_SUB:
			{
				log_print(LOG_DEBUG, "%04d: SUB\n", pos);
				if (vartype[vartypecount - 1] != vartype[vartypecount - 2])
				{
					module->program[module->proglen++] = 0x58; // pop %eax
					#ifdef __amd64__
					amd64_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					#ifdef __i386__
					x86_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					module->program[module->proglen++] = 0x50; // push %eax
				}
				#ifdef __amd64__
				amd64_sub(module, REG_RAX, REG_RCX, vartype[vartypecount - 2]);
				#endif
				#ifdef __i386__
				x86_sub(module, REG_EAX, REG_ECX, vartype[vartypecount - 2]);
				#endif
				
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_MUL:
			{
				log_print(LOG_DEBUG, "%04d: MUL\n", pos);
				if (vartype[vartypecount - 1] != vartype[vartypecount - 2])
				{
					module->program[module->proglen++] = 0x58; // pop %eax
					#ifdef __amd64__
					amd64_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					#ifdef __i386__
					x86_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					module->program[module->proglen++] = 0x50; // push %eax
				}
				#ifdef __amd64__
				amd64_mul(module, REG_RAX, REG_RCX, vartype[vartypecount - 2]);
				#endif
				#ifdef __i386__
				x86_mul(module, REG_EAX, REG_ECX, vartype[vartypecount - 2]);
				#endif
				
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_DIV:
			{
				log_print(LOG_DEBUG, "%04d: DIV\n", pos);
				if (vartype[vartypecount - 1] != vartype[vartypecount - 2])
				{
					module->program[module->proglen++] = 0x58; // pop %eax
					#ifdef __amd64__
					amd64_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					#ifdef __i386__
					x86_convert(module, vartype[vartypecount - 1], vartype[vartypecount - 2]);
					#endif
					module->program[module->proglen++] = 0x50; // push %eax
				}
				#ifdef __amd64__
				amd64_div(module, REG_RAX, REG_RCX, vartype[vartypecount - 2]);
				#endif
				#ifdef __i386__
				x86_div(module, REG_EAX, REG_ECX, vartype[vartypecount - 2]);
				#endif
				
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_AND:
			{
				log_print(LOG_DEBUG, "%04d: AND\n", pos);
				
				#ifdef __amd64__
				amd64_and(module, REG_RAX, REG_RCX);
				#endif
				#ifdef __i386__
				x86_and(module, REG_EAX, REG_ECX);
				#endif
								
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_OR:
			{
				log_print(LOG_DEBUG, "%04d: OR\n", pos);
				
				#ifdef __amd64__
				amd64_or(module, REG_RAX, REG_RCX);
				#endif
				#ifdef __i386__
				x86_or(module, REG_EAX, REG_ECX);
				#endif
								
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_XOR:
			{
				log_print(LOG_DEBUG, "%04d: XOR\n", pos);
				
				#ifdef __amd64__
				amd64_xor(module, REG_RAX, REG_RCX);
				#endif
				#ifdef __i386__
				x86_xor(module, REG_EAX, REG_ECX);
				#endif
								
				vartypecount--;
				
				pos += 1;
				break;
			}
			case CMD_CONCAT:
			{
				log_print(LOG_DEBUG, "%04d: CONCAT\n", pos);
				pos += 1;
				break;
			}
			case CMD_CALL:
			{
				int funcname = *((int*)(&module->bytecode[pos + 1]));
				
				log_print(LOG_DEBUG, "%04d: CALL (%s, %d arguments, retval: %d)\n", pos, module->strings[funcname],
					module->bytecode[pos + 5], module->bytecode[pos + 6]);
				
				void *funcaddr = vm_get_function_address(vm, module->strings[funcname]);
				if (!funcaddr)
				{
					log_print(LOG_ERROR, "Error: Function \"%s\" not found!\n", module->strings[funcname]);
					while(1);
				}
				
				int argcount = -1;
				int argtypes[15];
				int j;
				int rettype = VAR_TYPE_UNKNOWN;
				#ifdef __amd64__
				int intargcount = 0;
				int floatargcount = 0;
				#endif
				// External function?
				for (i = 0; i < vm->funccount; i++)
				{
					if (!strcmp(module->strings[funcname], vm->functions[i].name))
					{
						argcount = vm->functions[i].argcount;
						memcpy(argtypes, vm->functions[i].argtypes, argcount * sizeof(int));
						rettype = vm->functions[i].returntype;
						break;
					}
				}
				if (argcount == -1)
				{
					// Internal function?
					int modidx = -1;
					int funcidx = -1;
					// Look for function
					for (i = 0; i < vm->modcount; i++)
					{
						for (j = 0; j < vm->modules[i]->funccount; j++)
						{
							if (funcname == vm->modules[i]->functions[j].name)
							{
								modidx = i;
								funcidx = j;
								break;
							}
						}
					}
					// Get argument types
					argcount = vm->modules[modidx]->functions[funcidx].argcount;
					rettype = vm->modules[modidx]->functions[funcidx].rettype;
					for (i = 0; i < argcount; i++)
					{
						for (j = 0; j < vm->modules[modidx]->functions[funcidx].varcount; j++)
						{
							if (vm->modules[modidx]->functions[funcidx].argnames[i] == vm->modules[modidx]->functions[funcidx].vars[i].name)
							{
								argtypes[i] = vm->modules[modidx]->functions[funcidx].vars[i].type;
								break;
							}
						}
					}
				}
				#ifdef __amd64__
				for (i = 0; i < argcount; i++)
				{
					if (argtypes[i] == VAR_TYPE_FLOAT)
					{
						floatargcount++;
					}
					else
					{
						intargcount++;
					}
				}
				#endif
				if (module->bytecode[pos + 5] > argcount)
				{
					log_print(LOG_WARNING, "Warning: Too many arguments for function \"%s\".\n", module->strings[funcname]);
				}
				if (module->bytecode[pos + 5] < argcount)
				{
					log_print(LOG_WARNING, "Warning: Too few arguments for function \"%s\".\n", module->strings[funcname]);
					argcount = module->bytecode[pos + 5];
					#ifdef __amd64__
					for (i = 0; i < argcount; i++)
					{
						if (argtypes[i] == VAR_TYPE_FLOAT)
						{
							floatargcount++;
						}
						else
						{
							intargcount++;
						}
					}
					#endif
				}
				// Push arguments
				#ifdef __amd64__
				for (i = module->bytecode[pos + 5] - 1; i >= (intargcount + floatargcount); i--)
				{
					if (vartype[vartypecount - 1] != VAR_TYPE_INT)
					{
						module->program[module->proglen++] = 0x58; // pop %eax
						amd64_convert(module, vartype[vartypecount - 1], VAR_TYPE_INT);
						module->program[module->proglen++] = 0x50; // push %eax
					}
					amd64_set_int_argument(module, i - floatargcount);
					vartypecount--;
				}
				for (i = intargcount + floatargcount - 1; i >= 0; i--)
				{
					if (vartype[vartypecount - 1] != argtypes[i])
					{
						module->program[module->proglen++] = 0x58; // pop %eax
						amd64_convert(module, vartype[vartypecount - 1], argtypes[i]);
						module->program[module->proglen++] = 0x50; // push %eax
					}
					if (argtypes[i] == VAR_TYPE_FLOAT)
					{
						floatargcount--;
						amd64_set_float_argument(module, floatargcount);
					}
					else
					{
						intargcount--;
						amd64_set_int_argument(module, intargcount);
					}
					vartypecount--;
				}
				#endif
				#ifdef __i386__
				x86_flip_args(module, module->bytecode[pos + 5], 0, 1, &vartype[vartypecount - module->bytecode[pos + 5]], argtypes);
				vartypecount -= module->bytecode[pos + 5];
				#endif
				
				// Call function
				#ifdef __amd64__
				if (!vartypecount % 2)
				{
					module->program[module->proglen++] = 0x51; // push %ecx
				}
				amd64_call_function(module, funcaddr);
				if (!vartypecount % 2)
				{
					module->program[module->proglen++] = 0x59; // pop %ecx
				}
				#endif
				#ifdef __i386__
				x86_call_function(module, funcaddr);
				x86_adjust_esp(module, module->bytecode[pos + 5] * 4);
				#endif

				// Return value
				if (module->bytecode[pos + 6])
				{
					module->program[module->proglen++] = 0x50; // push %eax
					if (rettype == VAR_TYPE_FLOAT)
					{
						#ifdef __amd64__
						module->program[module->proglen++] = 0xf3; // mov %xmm0, (%rsp)
						module->program[module->proglen++] = 0x0f;
						module->program[module->proglen++] = 0x11;
						module->program[module->proglen++] = 0x04;
						module->program[module->proglen++] = 0x24;
						#endif
						#ifdef __i386__
						module->program[module->proglen++] = 0xd9; // fstp (%esp)
						module->program[module->proglen++] = 0x1c;
						module->program[module->proglen++] = 0x24;
						#endif
					}
					vartypecount++;
					vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
					vartype[vartypecount - 1] = rettype;
				}

				pos += 7;
				break;
			}
			case CMD_CALLMEMBER:
			{
				int var = *((int*)(&module->bytecode[pos + 1]));
				log_print(LOG_DEBUG, "%04d: CALLMEMBER (%s, %d arguments)\n", pos, module->strings[var], module->bytecode[pos + 5]);
				
				struct_info_t *info = vm_get_struct_info(vm, currenttype);
				if (!info)
				{
					log_print(LOG_ERROR, "Error: Structure info for \"%s\" not found.\n", currenttype);
					while(1);
				}
				void *funcaddr = 0;
				int argcount = 0;
				int argtypes[15];
				int rettype = 0;
				argtypes[0] = VAR_TYPE_USERPTR;
				#ifdef __amd64__
				int intargcount = 0;
				int floatargcount = 0;
				#endif
				for (i = 0; i < info->funccount; i++)
				{
					if (!strcmp(module->strings[var], info->functions[i].name))
					{
						funcaddr = info->functions[i].function;
						#ifdef __amd64__
						argcount = info->functions[i].argcount + 1;
						memcpy(&argtypes[1], info->functions[i].argtypes, (argcount - 1) * sizeof(int));
						#endif
						#ifdef __i386__
						argcount = info->functions[i].argcount;
						memcpy(argtypes, info->functions[i].argtypes, argcount * sizeof(int));
						#endif
						rettype = info->functions[i].returntype;
						break;
					}
				}
				if (!funcaddr)
				{
					log_print(LOG_ERROR, "Error: Member function \"%s\" not found.\n", module->strings[var]);
					while(1);
				}
				#ifdef __amd64__
				module->bytecode[pos + 5] += 1;
				for (i = 1; i < argcount; i++)
				{
					if (argtypes[i] == VAR_TYPE_FLOAT)
					{
						floatargcount++;
					}
					else
					{
						intargcount++;
					}
				}
				intargcount++;
				#endif
				if (module->bytecode[pos + 5] > argcount)
				{
					log_print(LOG_WARNING, "Warning: Too many arguments for function \"%s\".\n", module->strings[var]);
				}
				if (module->bytecode[pos + 5] < argcount)
				{
					log_print(LOG_WARNING, "Warning: Too few arguments for function \"%s\".\n", module->strings[var]);
					argcount = module->bytecode[pos + 5];
					#ifdef __amd64__
					for (i = 0; i < argcount; i++)
					{
						if (argtypes[i] == VAR_TYPE_FLOAT)
						{
							floatargcount++;
						}
						else
						{
							intargcount++;
						}
					}
					#endif
				}
				// Push arguments
				#ifdef __amd64__
				amd64_set_int_argument(module, 0);
				vartypecount--;
				for (i = module->bytecode[pos + 5] - 1; i >= (intargcount + floatargcount); i--)
				{
					if (vartype[vartypecount - 1] != VAR_TYPE_INT)
					{
						module->program[module->proglen++] = 0x58; // pop %eax
						amd64_convert(module, vartype[vartypecount - 1], VAR_TYPE_INT);
						module->program[module->proglen++] = 0x50; // push %eax
					}
					amd64_set_int_argument(module, i - floatargcount);
					vartypecount--;
				}
				for (i = intargcount + floatargcount - 1; i >= 1; i--)
				{
					if (vartype[vartypecount - 1] != argtypes[i])
					{
						module->program[module->proglen++] = 0x58; // pop %eax
						amd64_convert(module, vartype[vartypecount - 1], argtypes[i]);
						module->program[module->proglen++] = 0x50; // push %eax
					}
					if (argtypes[i] == VAR_TYPE_FLOAT)
					{
						floatargcount--;
						amd64_set_float_argument(module, floatargcount);
					}
					else
					{
						intargcount--;
						amd64_set_int_argument(module, intargcount);
					}
					vartypecount--;
				}
				#endif
				#ifdef __i386__
				vartypecount--;
				x86_flip_args(module, module->bytecode[pos + 5], 1, 1, &vartype[vartypecount - module->bytecode[pos + 5]], argtypes);
				vartypecount -= module->bytecode[pos + 5];
				#endif
				
				// Call function
				#ifdef __amd64__
				if (!vartypecount % 2)
				{
					module->program[module->proglen++] = 0x51; // push %ecx
				}
				amd64_call_function(module, funcaddr);
				if (!vartypecount % 2)
				{
					module->program[module->proglen++] = 0x59; // pop %ecx
				}
				#endif
				#ifdef __i386__
				x86_call_function(module, funcaddr);
				x86_adjust_esp(module, module->bytecode[pos + 5] * 4);
				#endif

				// Return value
				if (module->bytecode[pos + 6])
				{
					module->program[module->proglen++] = 0x50; // push %eax
					if (rettype == VAR_TYPE_FLOAT)
					{
						#ifdef __amd64__
						module->program[module->proglen++] = 0xf3; // mov %xmm0, (%rsp)
						module->program[module->proglen++] = 0x0f;
						module->program[module->proglen++] = 0x11;
						module->program[module->proglen++] = 0x04;
						module->program[module->proglen++] = 0x24;
						#endif
						#ifdef __i386__
						module->program[module->proglen++] = 0xd9; // fstp (%esp)
						module->program[module->proglen++] = 0x1c;
						module->program[module->proglen++] = 0x24;
						#endif
					}
					vartypecount++;
					vartype = (int*)realloc(vartype, vartypecount * sizeof(int));
					vartype[vartypecount - 1] = rettype;
				}
				
				pos += 7;
				break;
			}

			case CMD_END:
			{
				log_print(LOG_DEBUG, "%04d: END\n", pos);
				pos += 1;
				break;
			}
			case CMD_RETURN:
			{
				log_print(LOG_DEBUG, "%04d: RETURN (%d)\n", pos, module->bytecode[pos + 1]);
				
				if (module->bytecode[pos + 1])
				{
					// Get return value
					if (func->rettype == VAR_TYPE_FLOAT)
					{
						#ifdef __amd64__
						module->program[module->proglen++] = 0xf3; // mov (%rsp), %xmm0
						module->program[module->proglen++] = 0x0f;
						module->program[module->proglen++] = 0x10;
						module->program[module->proglen++] = 0x04;
						module->program[module->proglen++] = 0x24;
						#endif
						#ifdef __i386__
						module->program[module->proglen++] = 0xd9; // flds (%esp)
						module->program[module->proglen++] = 0x04;
						module->program[module->proglen++] = 0x24;
						#endif

					}
					module->program[module->proglen++] = 0x58; // pop %eax
				}
				
				module->program[module->proglen++] = 0xe9;
				*((int*)&module->program[module->proglen]) = 0;
				module->proglen += 4;
				// Adding return to unresolved list
				jmpaddr = (char**)realloc(jmpaddr, sizeof(char*) * (unrescount + 1));
				jmpaddr[unrescount] = &module->program[module->proglen - 4];
				jmplabel = (int*)realloc(jmplabel, sizeof(int) * (unrescount + 1));
				jmplabel[unrescount] = -1;
				unrescount++;
				
				pos += 2;
				break;
			}
			default:
				log_print(LOG_ERROR, "%04d: Unknown bytecode!\n", pos);
				pos++;
		}
	}
	
	// Resolve return jumps
	for (i = 0; i < unrescount; i++)
	{
		if (jmplabel[i] == -1)
		{
			#ifdef __amd64__
			long long addrdiff = &module->program[module->proglen] - (jmpaddr[i] + 4);
			*((int*)jmpaddr[i]) = addrdiff;
			#endif
			#ifdef __i386__
			int addrdiff = &module->program[module->proglen] - (jmpaddr[i] + 4);
			*((int*)jmpaddr[i]) = addrdiff;
			#endif
		}
	}
	
	// Free space and call destructors
	module->program[module->proglen++] = 0x50; // push %eax
	for (i = 0; i < func->varcount; i++)
	{
		if (func->vars[i].type == VAR_TYPE_STRUCT)
		{
			struct_info_t *info = vm_get_struct_info(vm, module->strings[func->vars[i].type_name]);
			if (info)
			{
				// Call destructor
				if (info->destructor)
				{
					#ifdef __amd64__
					amd64_get_local_var(module, i, REG_RDI);
					amd64_call_function(module, (void*)info->destructor);
					#endif
					#ifdef __i386__
					module->program[module->proglen++] = 0xff; // push -i * 4(%ebp)
					module->program[module->proglen++] = 0x75;
					module->program[module->proglen++] = 0x100 - ((i + 1) * 4);
					x86_call_function(module, (void*)info->destructor);
					x86_adjust_esp(module, 4);
					#endif
				}
				// Free memory
				#ifdef __amd64__
				amd64_get_local_var(module, i, REG_RDI);
				amd64_call_function(module, (void*)free);
				#endif
				#ifdef __i386__
				module->program[module->proglen++] = 0xff; // push -i * 4(%ebp)
				module->program[module->proglen++] = 0x75;
				module->program[module->proglen++] = 0x100 - ((i + 1) * 4);
				x86_call_function(module, (void*)free);
				x86_adjust_esp(module, 4);
				#endif
			}
		}
	}
	
	module->program[module->proglen++] = 0x58; // pop %eax
	// Free space for local variables
	#ifdef __amd64__
	module->program[module->proglen++] = 0x48; // add $func->varcount, %rsp
	module->program[module->proglen++] = 0x81;
	module->program[module->proglen++] = 0xc4;
	*((int*)&module->program[module->proglen]) = func->varcount * 8;
	module->proglen += sizeof(int);
	#endif
	#ifdef __i386__
	x86_adjust_esp(module, func->varcount * 4);
	#endif
	// Leave function
	#if defined(__amd64__) || defined(__i386__)
	module->program[module->proglen++] = 0xc9; // leave
	module->program[module->proglen++] = 0xc3; // ret
	#endif
	
	free(jmpaddr);
	free(jmplabel);
	free(labeladdr);
	free(labelidx);
	
	free(vartype);
	
	// Modify call stub
	#ifdef __amd64__
	long long addrdiff = &module->program[func->progoffset] - (func->call_stub + 5);
	#endif
	#ifdef __i386__
	int addrdiff = &module->program[func->progoffset] - (func->call_stub + 5);
	#endif
	log_print(LOG_DEBUG, "addrdiff: %p\n", (void*)addrdiff);
	#if defined(__amd64__) || defined(__i386__)
	if ((addrdiff > 0x7FFFFFFF) || (addrdiff < -0x7FFFFFFF))
	{
		// TODO
		log_print(LOG_ERROR, "Difference too long.\n");
		while(1);
	}
	func->call_stub[0] = 0xe9;
	*((int*)&func->call_stub[1]) = addrdiff;
	#endif
	log_print(LOG_WARNING, "Code size: %d bytes.\n", module->proglen - func->progoffset);
}

