
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "module.h"
#include "bytecode.h"
#include "parser.h"
#include "vm.h"
#include "logger.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef __unix__
#include <sys/mman.h>
#endif

int save_string(module_t *module, char *string)
{
	// Do not save duplicated strings
	// TODO: This is slow.
	int i;
	for (i = 0; i < module->stringcount; i++)
	{
		if (!strcmp(module->strings[i], string)) return i;
	}
	
	module->strings = (char**)realloc(module->strings, sizeof(char*) * (module->stringcount + 1));
	module->strings[module->stringcount] = strdup(string);
	module->stringcount++;
	return module->stringcount - 1;
}
int get_string(module_t *module, const char *string)
{
	int i;
	for (i = 0; i < module->stringcount; i++)
	{
		if (!strcmp(module->strings[i], string)) return i;
	}
	return 0;
}
int get_variable_index(mod_func_t *function, int name)
{
	int i;
	for (i = 0; i < function->varcount; i++)
	{
		if (function->vars[i].name == name)
		{
			return i;
		}
	}
	return -1;
}

int get_type(struct ps_vm_t *vm, module_t *module, char *type_name)
{
	// Base variable types
	if (!strcmp(type_name, "int"))
	{
		return VAR_TYPE_INT;
	}
	else if (!strcmp(type_name, "float"))
	{
		return VAR_TYPE_FLOAT;
	}
	else
	{
		// Search for struct
		int i;
		for (i = 0; i < vm->structcount; i++)
		{
			if (!strcmp(type_name, vm->structs[i]->name))
			{
				return VAR_TYPE_STRUCT;
			}
		}
		return VAR_TYPE_UNKNOWN;
	}
}
int var_is_valid(module_t *module, mod_func_t *function, int name)
{
	int i;
	for (i = 0; i < module->varcount; i++)
	{
		if (module->vars[i].name == name)
		{
			return 1;
		}
	}
	for (i = 0; i < function->varcount; i++)
	{
		if (function->vars[i].name == name)
		{
			return 1;
		}
	}
	return 0;
}

char buffer[1024];
int length = -1;

#define OP_PLUS 1
#define OP_MINUS 2
#define OP_MUL 3
#define OP_DIV 4
#define OP_LBRACKET 5
#define OP_RBRACKET 6
#define OP_CONCAT 7
#define OP_EQUAL 8
#define OP_NEQUAL 9
#define OP_MORE 10
#define OP_LESS 11
#define OP_MOREEQUAL 12
#define OP_LESSEQUAL 13
#define OP_MEMBER 14
#define OP_AND 15
#define OP_OR 16
#define OP_XOR 17

int parse_expression_part(module_t *module, mod_func_t *function, FILE *file, int closingbracket)
{
	int expectingvar = 1;
	int expectingfunction = 0;
	int expectingmember = 0;
	char *functionname = 0;
	int membercount = 0;
	int *membername;
	
	int operators[256];
	int opcount = 0;
	
	int i;
	
	while (strcmp(buffer, ",") && strcmp(buffer, ";") && strcmp(buffer, "then")
		&& strcmp(buffer, "to") && strcmp(buffer, "step") && strcmp(buffer, "do"))
	{
		if (membercount > 0)
		{
			if (!strcmp(buffer, "("))
			{
				// Member function called!
				// Read arguments
				length = read_word(file, buffer);
				if (length == -1)
				{
					log_print(LOG_ERROR, "Error: Expected ')', got \"%s\".\n", buffer);
					return 0;
				}
				int argcount = 0;
				while (strcmp(buffer, ")"))
				{
					argcount++;
					if (!parse_expression_part(module, function, file, 1))
					{
						return 0;
					}
					if (!strcmp(buffer, ")"))
					{
						break;
					}
					length = read_word(file, buffer);
					if (length == -1)
					{
						log_print(LOG_ERROR, "Error: Unexpected EOF.\n");
						return 0;
					}
				}
				length = read_word(file, buffer);
				if (length == -1)
				{
					log_print(LOG_ERROR, "Error: Expected ')', got \"%s\".\n", buffer);
					return 0;
				}
				// Call function
				bytecode_add_get(module, membername[0]);
				for (i = 1; i < membercount - 1; i++)
				{
					bytecode_add_getmember(module, membername[i]);
				}
				bytecode_add_callmember(module, membername[membercount - 1], argcount, 1);
				membercount = 0;
				free(membername);
				if (!strcmp(buffer, ";"))
				{
					break;
				}
			}
			else if (!strcmp(buffer, "."))
			{
				// Member accessed
				length = read_word(file, buffer);
				if ((length == -1) || !isalpha(buffer[0]))
				{
					log_print(LOG_ERROR, "Error: Expected member name.\n");
					return 0;
				}
				// Save member
				int name = save_string(module, buffer);
				membercount++;
				membername = (int*)realloc(membername, membercount * sizeof(int));
				membername[membercount - 1] = name;
			}
			else
			{
				log_print(LOG_DEBUG, "Getting member.\n");
				// Get member
				bytecode_add_get(module, membername[0]);
				for (i = 1; i < membercount; i++)
				{
					bytecode_add_getmember(module, membername[i]);
				}
				membercount = 0;
				free(membername);
			}
			expectingvar = 0;
		}
		
		if (expectingfunction)
		{
			// Function 
			expectingfunction = 0;
			if (strcmp(buffer, "("))
			{
				log_print(LOG_ERROR, "Error: Expected '('.\n");
				free(functionname);
				return 0;
			}
			int name = save_string(module, functionname);
			free(functionname);
			length = read_word(file, buffer);
			if (length == -1)
			{
				log_print(LOG_ERROR, "Error: Expected ')', got EOF.\n");
				return 0;
			}
			int argcount = 0;
			while (strcmp(buffer, ")"))
			{
				argcount++;
				if (!parse_expression_part(module, function, file, 1))
				{
					return 0;
				}
				if (!strcmp(buffer, ")"))
				{
					break;
				}
				length = read_word(file, buffer);
				if (length == -1)
				{
					log_print(LOG_ERROR, "Error: Unexpected EOF.\n");
					return 0;
				}
			}
			// Call function
			bytecode_add_call(module, name, argcount, 1);
			expectingvar = 0;
		}
		else if (expectingmember)
		{
			expectingmember = 0;
			if ((length == -1) || !isalpha(buffer[0]))
			{
				log_print(LOG_ERROR, "Error: Expected member, got \"%s\".\n", buffer);
				return 0;
			}
			
			// Remove CMD_GET from code and store variable name
			module->codelen -= 4;
			int varname = *((int*)(module->bytecode + module->codelen));
			module->codelen--;
			
			// Add member to list
			membercount = 2;
			membername = (int*)malloc(sizeof(int) * 2);
			membername[0] = varname;
			membername[1] = save_string(module, buffer);
		}
		else if (expectingvar)
		{
			// Operators which do not need 2 values
			if (!strcmp(buffer, "-"))
			{
				// TODO: Floats?
				bytecode_add_getc(module, 0);
				operators[opcount++] = OP_MINUS;
			}
			else if (!strcmp(buffer, "+"))
			{
				// TODO: Floats?
				bytecode_add_getc(module, 0);
				operators[opcount++] = OP_PLUS;
			}
			else if (!strcmp(buffer, "("))
			{
				operators[opcount++] = OP_LBRACKET;
			}
			else if (isalpha(buffer[0]))
			{
				// Variable
				int var = get_string(module, buffer);
				if (!var || !var_is_valid(module, function, var))
				{
					log_print(LOG_DEBUG, "Variable not valid: \"%s\"\n", buffer);
					functionname = strdup(buffer);
					expectingfunction = 1;
				}
				else
				{
					bytecode_add_get(module, var);
					expectingvar = 0;
				}
			}
			else if (isdigit(buffer[0]))
			{
				// Constant value
				if (strchr(buffer, '.'))
				{
					bytecode_add_getf(module, atof(buffer));
				}
				else
				{
					bytecode_add_getc(module, atoi(buffer));
				}
				expectingvar = 0;
			}
			else if (buffer[0] == '\"')
			{
				buffer[strlen(buffer) - 1] = 0;
				int str = save_string(module, buffer + 1);
				
				// Constant string
				bytecode_add_gets(module, str);
				expectingvar = 0;
			}
			else
			{
				log_print(LOG_ERROR, "Error: Unknown word (\"%s\").\n", buffer);
				return 0;
			}
		}
		else
		{
			// Operator expected
			if (!strcmp(buffer, "-"))
			{
				if (opcount > 0)
				{
					if (operators[opcount - 1] == OP_MUL)
					{
						module->bytecode[module->codelen++] = CMD_MUL;
						opcount--;
					}
					else if (operators[opcount - 1] == OP_DIV)
					{
						module->bytecode[module->codelen++] = CMD_DIV;
						opcount--;
					}
				}
				operators[opcount++] = OP_MINUS;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "+"))
			{
				if (opcount > 0)
				{
					if (operators[opcount - 1] == OP_MUL)
					{
						module->bytecode[module->codelen++] = CMD_MUL;
						opcount--;
					}
					else if (operators[opcount - 1] == OP_DIV)
					{
						module->bytecode[module->codelen++] = CMD_DIV;
						opcount--;
					}
				}
				operators[opcount++] = OP_PLUS;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "*"))
			{
				operators[opcount++] = OP_MUL;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "/"))
			{
				operators[opcount++] = OP_DIV;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, ".."))
			{
				operators[opcount++] = OP_CONCAT;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "=="))
			{
				// TODO: Operator priority
				operators[opcount++] = OP_EQUAL;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "!="))
			{
				operators[opcount++] = OP_NEQUAL;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "<"))
			{
				operators[opcount++] = OP_LESS;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, ">"))
			{
				operators[opcount++] = OP_MORE;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "<="))
			{
				operators[opcount++] = OP_LESSEQUAL;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, ">="))
			{
				operators[opcount++] = OP_MOREEQUAL;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "and"))
			{
				operators[opcount++] = OP_AND;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "or"))
			{
				operators[opcount++] = OP_OR;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, "xor"))
			{
				operators[opcount++] = OP_XOR;
				expectingvar = 1;
			}
			else if (!strcmp(buffer, ")"))
			{
				// Add operators after the left bracket
				if (opcount == 0)
				{
					if (closingbracket)
					{
						return 1;
					}
					log_print(LOG_ERROR, "Error: ')' without '('.\n");
					return 0;
				}
				while (operators[opcount - 1] != OP_LBRACKET)
				{
					if (operators[opcount - 1] == OP_MINUS)
					{
						module->bytecode[module->codelen++] = CMD_SUB;
					}
					else if (operators[opcount - 1] == OP_PLUS)
					{
						module->bytecode[module->codelen++] = CMD_ADD;
					}
					else if (operators[opcount - 1] == OP_MUL)
					{
						module->bytecode[module->codelen++] = CMD_MUL;
					}
					else if (operators[opcount - 1] == OP_DIV)
					{
						module->bytecode[module->codelen++] = CMD_DIV;
					}
					else if (operators[opcount - 1] == OP_CONCAT)
					{
						module->bytecode[module->codelen++] = CMD_CONCAT;
					}
					else if (operators[opcount - 1] == OP_EQUAL)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_EQUAL;
					}
					else if (operators[opcount - 1] == OP_NEQUAL)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_GREATER | CMP_SMALLER;
					}
					else if (operators[opcount - 1] == OP_MORE)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_GREATER;
					}
					else if (operators[opcount - 1] == OP_LESS)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_SMALLER;
					}
					else if (operators[opcount - 1] == OP_MOREEQUAL)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_GREATER | CMP_EQUAL;
					}
					else if (operators[opcount - 1] == OP_LESSEQUAL)
					{
						module->bytecode[module->codelen++] = CMD_CMP;
						module->bytecode[module->codelen++] = CMP_EQUAL | CMP_SMALLER;
					}
					else if (operators[opcount - 1] == OP_AND)
					{
						module->bytecode[module->codelen++] = CMD_AND;
					}
					else if (operators[opcount - 1] == OP_OR)
					{
						module->bytecode[module->codelen++] = CMD_OR;
					}
					else if (operators[opcount - 1] == OP_XOR)
					{
						module->bytecode[module->codelen++] = CMD_XOR;
					}
					opcount--;
					if (opcount == 0)
					{
						if (closingbracket)
						{
							return 1;
						}
						log_print(LOG_ERROR, "Error: ')' without '('.\n");
						return 0;
					}
				}
				// Delete left bracket
				opcount--;
				
				expectingvar = 0;
			}
			else if (!strcmp(buffer, "("))
			{
				// Function call
				expectingfunction = 1;
			}
			else if (!strcmp(buffer, "."))
			{
				expectingmember = 1;
			}
			else
			{
				log_print(LOG_ERROR, "Error: Expected operator, got \"%s\".\n", buffer);
				return 0;
			}
		}
		
		length = read_word(file, buffer);
		if (length == -1)
		{
			log_print(LOG_ERROR, "Error: EOF while reading expression.\n");
			return 0;
		}
	}
	
	// Get remaining members
	if (membercount > 0)
	{
		bytecode_add_get(module, membername[0]);
		for (i = 1; i < membercount; i++)
		{
			bytecode_add_getmember(module, membername[1]);
		}
		membercount = 0;
		free(membername);
	}
	
	// Add all operators
	while (opcount > 0)
	{
		if (operators[opcount - 1] == OP_MINUS)
		{
			module->bytecode[module->codelen++] = CMD_SUB;
		}
		else if (operators[opcount - 1] == OP_PLUS)
		{
			module->bytecode[module->codelen++] = CMD_ADD;
		}
		else if (operators[opcount - 1] == OP_MUL)
		{
			module->bytecode[module->codelen++] = CMD_MUL;
		}
		else if (operators[opcount - 1] == OP_DIV)
		{
			module->bytecode[module->codelen++] = CMD_DIV;
		}
		else if (operators[opcount - 1] == OP_CONCAT)
		{
			module->bytecode[module->codelen++] = CMD_CONCAT;
		}
		else if (operators[opcount - 1] == OP_EQUAL)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_EQUAL;
		}
		else if (operators[opcount - 1] == OP_NEQUAL)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_GREATER | CMP_SMALLER;
		}
		else if (operators[opcount - 1] == OP_MORE)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_GREATER;
		}
		else if (operators[opcount - 1] == OP_LESS)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_SMALLER;
		}
		else if (operators[opcount - 1] == OP_MOREEQUAL)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_GREATER | CMP_EQUAL;
		}
		else if (operators[opcount - 1] == OP_LESSEQUAL)
		{
			module->bytecode[module->codelen++] = CMD_CMP;
			module->bytecode[module->codelen++] = CMP_EQUAL | CMP_SMALLER;
		}
		else if (operators[opcount - 1] == OP_AND)
		{
			module->bytecode[module->codelen++] = CMD_AND;
		}
		else if (operators[opcount - 1] == OP_OR)
		{
			module->bytecode[module->codelen++] = CMD_OR;
		}
		else if (operators[opcount - 1] == OP_XOR)
		{
			module->bytecode[module->codelen++] = CMD_XOR;
		}
		else if (operators[opcount - 1] == OP_LBRACKET)
		{
			log_print(LOG_ERROR, "Error: Expected ')'.\n");
			return 0;
		}
		opcount--;
	}
	
	return 1;
}

int parse_expression(struct ps_vm_t *vm, module_t *module, FILE *file, mod_func_t *function)
{
	char *firstword = strdup(buffer);
	char *varname = firstword;
	
	// Read next word to determine type
	length = read_word(file, buffer);
	if (length == -1)
	{
		log_print(LOG_ERROR, "Error: Unexpected EOF.\n");
		free(firstword);
		return 0;
	}
	if (isalpha(buffer[0]))
	{
		// Second word is a name, so this is a variable declaration.
		char *vartype = firstword;
		varname = strdup(buffer);
		
		// Add local variable
		mod_var_t newvar;
		newvar.type_name = save_string(module, vartype);
		newvar.name = save_string(module, varname);
		newvar.value = 0;
		newvar.type = get_type(vm, module, vartype);
		newvar.addr = 0;
		function->vars = (mod_var_t*)realloc(function->vars, sizeof(mod_var_t) * (function->varcount + 1));
		function->vars[function->varcount] = newvar;
		function->varcount++;
		//printf("Variable: %s/%s.\n", varname, vartype);
		free(vartype);
		
		length = read_word(file, buffer);
		if (length == -1)
		{
			log_print(LOG_ERROR, "Error: Unexpected EOF after variable definition.\n");
			free(varname);
			return 0;
		}
		//printf("Local variable.\n");
		if (!strcmp(buffer, ";"))
		{
			free(varname);
			return 1;
		}
		//printf("Buffer: \"%s\"\n", buffer);
	}
	if (!strcmp(buffer, "("))
	{
		//printf("Function call.\n");
		
		int argcount = 0;
		// Read arguments
		length = read_word(file, buffer);
		if (strcmp(buffer, ")"))
		{
			while (strcmp(buffer, ")"))
			{
				if (!parse_expression_part(module, function, file, 1))
				{
					log_print(LOG_ERROR, "Error: Could not parse function argument.");
					free(varname);
					return 0;
				}
				
				if (strcmp(buffer, ",") && strcmp(buffer, ")"))
				{
					log_print(LOG_ERROR, "Error: Expected ',' or ')' after function argument.\n");
					log_print(LOG_ERROR, "Got: \"%s\"\n", buffer);
					free(varname);
					return 0;
				}
				argcount++;
				if (strcmp(buffer, ")"))
				{
					length = read_word(file, buffer);
				}
			}
		}
		
		length = read_word(file, buffer);
		if (strcmp(buffer, ";"))
		{
			log_print(LOG_ERROR, "Error: Expected ';'.\n");
			free(varname);
			return 0;
		}
		
		// Save function
		bytecode_add_call(module, save_string(module, varname), argcount, 0);
	}
	else if (!strcmp(buffer, "="))
	{
		length = read_word(file, buffer);
		if (!parse_expression_part(module, function, file, 0))
		{
			free(varname);
			return 0;
		}
		int var = get_string(module, varname);
		if (var == 0)
		{
			log_print(LOG_ERROR, "Error: Undeclared variable \"%s\".\n", varname);
			free(varname);
			return 0;
		}
		bytecode_add_set(module, var);
	}
	else if (!strcmp(buffer, "+="))
	{
		length = read_word(file, buffer);
		if (!parse_expression_part(module, function, file, 0))
		{
			free(varname);
			return 0;
		}
		int var = get_string(module, varname);
		if (var == 0)
		{
			free(varname);
			log_print(LOG_ERROR, "Error: Undeclared variable \"%s\".\n", varname);
			return 0;
		}
		bytecode_add_get(module, var);
		module->bytecode[module->codelen++] = CMD_ADD;
		bytecode_add_set(module, var);
	}
	else if (!strcmp(buffer, "-="))
	{
		length = read_word(file, buffer);
		if (!parse_expression_part(module, function, file, 0))
		{
			free(varname);
			return 0;
		}
		int var = get_string(module, varname);
		if (var == 0)
		{
			log_print(LOG_ERROR, "Error: Undeclared variable \"%s\".\n", varname);
			free(varname);
			return 0;
		}
		bytecode_add_get(module, var);
		module->bytecode[module->codelen++] = CMD_SUB;
		bytecode_add_set(module, var);
	}
	else if (!strcmp(buffer, "."))
	{
		int var = get_string(module, varname);
		if (var == 0)
		{
			log_print(LOG_ERROR, "Error: Undeclared variable \"%s\".\n", varname);
			free(varname);
			return 0;
		}
		int membercount = 0;
		int *membername = 0;
		// Structure member
		while (1)
		{
			// Read member name
			length = read_word(file, buffer);
			if ((length == -1) || !isalpha(buffer[0]))
			{
				log_print(LOG_ERROR, "Error: Expected member name.\n");
				free(varname);
				return 0;
			}
			membercount++;
			membername = (int*)realloc(membername, sizeof(int) * membercount);
			membername[membercount - 1] = save_string(module, buffer);
			
			// Read operator afterwards
			length = read_word(file, buffer);
			if (length == -1)
			{
				log_print(LOG_ERROR, "Error: Unexpected EOF.\n");
				free(varname);
				return 0;
			}
			if (strcmp(buffer, "."))
			{
				break;
			}
		}
		if (!strcmp(buffer, "("))
		{
			// Call function
			int argcount = 0;
			// Read arguments
			length = read_word(file, buffer);
			if (strcmp(buffer, ")"))
			{
				while (strcmp(buffer, ")"))
				{
					if (!parse_expression_part(module, function, file, 1))
					{
						log_print(LOG_ERROR, "Error: Could not parse function argument.");
						free(varname);
						return 0;
					}
				
					if (strcmp(buffer, ",") && strcmp(buffer, ")"))
					{
						log_print(LOG_ERROR, "Error: Expected ',' or ')' after function argument.\n");
						log_print(LOG_ERROR, "Got: \"%s\"\n", buffer);
						free(varname);
						return 0;
					}
					argcount++;
					if (strcmp(buffer, ")"))
					{
						length = read_word(file, buffer);
					}
				}
			}
		
			length = read_word(file, buffer);
			if (strcmp(buffer, ";"))
			{
				log_print(LOG_ERROR, "Error: Expected ';'.\n");
				free(varname);
				return 0;
			}
		
			// Call member function
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_callmember(module, membername[membercount - 1], argcount, 0);
		}
		else if (!strcmp(buffer, "="))
		{
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(varname);
				return 0;
			}
			// Get member
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_setmember(module, membername[membercount - 1]);
		}
		else if (!strcmp(buffer, "+="))
		{
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(varname);
				return 0;
			}
			// Get member
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			module->bytecode[module->codelen++] = CMD_ADD;
			bytecode_add_get(module, var);
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_setmember(module, membername[membercount - 1]);
		}
		else if (!strcmp(buffer, "-="))
		{
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(varname);
				return 0;
			}
			// Get member
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			module->bytecode[module->codelen++] = CMD_SUB;
			bytecode_add_get(module, var);
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_setmember(module, membername[membercount - 1]);
		}
		else if (!strcmp(buffer, "*="))
		{
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(varname);
				return 0;
			}
			// Get member
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			module->bytecode[module->codelen++] = CMD_MUL;
			bytecode_add_get(module, var);
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_setmember(module, membername[membercount - 1]);
		}
		else if (!strcmp(buffer, "/="))
		{
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(varname);
				return 0;
			}
			// Get member
			bytecode_add_get(module, var);
			int i;
			for (i = 0; i < membercount; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			module->bytecode[module->codelen++] = CMD_DIV;
			bytecode_add_get(module, var);
			for (i = 0; i < membercount - 1; i++)
			{
				bytecode_add_getmember(module, membername[i]);
			}
			bytecode_add_setmember(module, membername[membercount - 1]);
		}
		else
		{
			log_print(LOG_ERROR, "Error: Expected '(' or '='.\n");
			free(varname);
			return 0;
		}
		free(membername);
	}
	else
	{
		log_print(LOG_ERROR, "Error: Unknown word (\"%s\").\n", buffer);
		free(varname);
		return 0;
	}
	//printf("Done.\n");
	free(varname);
	return 1;
}

int parse_function(struct ps_vm_t *vm, module_t *module, FILE *file)
{
	// Function name
	if (((length = read_word(file, buffer)) == -1) || (!isalpha(buffer[0])))
	{
		log_print(LOG_ERROR, "Error: Function name expected.\n");
		return 0;
	}
	char *funcname = strdup(buffer);
	char *rettype = 0;
	
	// Create function data
	module->functions = (mod_func_t*)realloc(module->functions, sizeof(mod_func_t) * (module->funccount + 1));
	mod_func_t *function = &module->functions[module->funccount];
	memset(function, 0, sizeof(mod_func_t));
	module->funccount++;
	
	function->name = save_string(module, funcname);
	free(funcname);
	function->bcoffset = module->codelen;
	
	// Arguments
	if (((length = read_word(file, buffer)) == -1) || strcmp(buffer, "("))
	{
		log_print(LOG_ERROR, "Error: '(' expected.\n");
		return 0;
	}
	
	while (((length = read_word(file, buffer)) != -1) && strcmp(buffer, ")"))
	{
		char *vartype = strdup(buffer);
		length = read_word(file, buffer);
		if (length != -1)
		{
			char *varname = strdup(buffer);
			int array = 0;
			
			length = read_word(file, buffer);
			if (!strcmp(buffer, "["))
			{
				length = read_word(file, buffer);
				if (strcmp(buffer, "]"))
				{
					log_print(LOG_ERROR, "Error: Need ']' after array declaration.\n");
					return 0;
				}
				else
				{
					array = 1;
					length = read_word(file, buffer);
				}
			}
			
			// Add argument
			mod_var_t newvar;
			newvar.type_name = save_string(module, vartype);
			newvar.name = save_string(module, varname);
			newvar.value = 0;
			newvar.type = get_type(vm, module, vartype);
			newvar.addr = 0;
			function->vars = (mod_var_t*)realloc(function->vars, sizeof(mod_var_t) * (function->varcount + 1));
			function->vars[function->varcount] = newvar;
			function->varcount++;
			free(vartype);
			free(varname);

			function->argnames = (int*)realloc(function->argnames, sizeof(int) * (function->argcount + 1));
			function->argnames[function->argcount] = newvar.name;
			function->argcount++;
			
			// Next argument?
			if (!strcmp(buffer, ")"))
			{
				break;
			}
			else if (strcmp(buffer, ","))
			{
				log_print(LOG_ERROR, "Error: Expected ',' or ')' after function argument.\n");
				return 0;
			}
		}
		else
		{
			free(vartype);
		}
	}
	if (length == -1)
	{
		log_print(LOG_ERROR, "Error: Could not read argument list.\n");
	}
	
	// Return value
	if (((length = read_word(file, buffer)) != -1) && !strcmp(buffer, ":"))
	{
		if (((length = read_word(file, buffer)) == -1) || (!isalpha(buffer[0])))
		{
			log_print(LOG_ERROR, "Error: Expected variable type after ':'.\n");
			free(funcname);
			return 0;
		}
		log_print(LOG_DEBUG, "Return type: \"%s\"\n", buffer);
		rettype = strdup(buffer);
		length = read_word(file, buffer);
		while (length == 0) length = read_word(file, buffer);
		function->rettypename = save_string(module, rettype);
		if (!strcmp(rettype, "float"))
		{
			function->rettype = VAR_TYPE_FLOAT;
		}
		else
		{
			function->rettype = VAR_TYPE_INT;
		}
		free(rettype);
	}
	if (length == -1)
	{
		log_print(LOG_ERROR, "Error: Could not read function header.\n");
		free(funcname);
		free(rettype);
		return 0;
	}
	
	int labelid = 0;
	ctrl_frame_t *ctrlstack = 0;
	int ctrlstacksize = 0;
	
	// Code
	while (length >= 0)
	{
		// Keywords
		if (!strcmp(buffer, "if"))
		{
			// If
			ctrlstack = (ctrl_frame_t*)realloc(ctrlstack, sizeof(ctrl_frame_t) * (ctrlstacksize + 1));
			ctrlstack[ctrlstacksize].id = labelid++;
			ctrlstack[ctrlstacksize].type = CTRL_FRAME_IF;
			ctrlstacksize++;
			
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(ctrlstack);
				return 0;
			}
			
			bytecode_add_if(module, ctrlstack[ctrlstacksize - 1].id);
		}
		else if (!strcmp(buffer, "elseif"))
		{
			// Else if
		}
		else if (!strcmp(buffer, "else"))
		{
			// Else
			if (ctrlstack[ctrlstacksize - 1].type != CTRL_FRAME_IF)
			{
				log_print(LOG_ERROR, "Error: else without if.\n");
				free(ctrlstack);
				return 0;
			}
			
			int iflabelid = ctrlstack[ctrlstacksize - 1].id;
			ctrlstack[ctrlstacksize - 1].type = CTRL_FRAME_ELSE;
			ctrlstack[ctrlstacksize - 1].id = labelid++;
			
			bytecode_add_jmp(module, ctrlstack[ctrlstacksize - 1].id);
			bytecode_add_label(module, iflabelid);
			log_print(LOG_DEBUG, "Else.\n");
		}
		else if (!strcmp(buffer, "for"))
		{
			// For
			ctrlstack = (ctrl_frame_t*)realloc(ctrlstack, sizeof(ctrl_frame_t) * (ctrlstacksize + 1));
			ctrlstack[ctrlstacksize].id = labelid;
			ctrlstack[ctrlstacksize].type = CTRL_FRAME_FOR;
			ctrlstacksize++;
			labelid += 2;
			
			// Read variable
			length = read_word(file, buffer);
			if ((length == -1) || !isalpha(buffer[0]))
			{
				log_print(LOG_ERROR, "Error: Expected variable name after for.\n");
				free(ctrlstack);
				return 0;
			}
			char *varname = strdup(buffer);
			length = read_word(file, buffer);
			if ((length == -1) || strcmp(buffer, "="))
			{
				log_print(LOG_ERROR, "Error: Expected '=' after variable name.\n");
				free(ctrlstack);
				free(varname);
				return 0;
			}
			length = read_word(file, buffer);
			
			if (!parse_expression_part(module, function, file, 0))
			{
				free(ctrlstack);
				free(varname);
				return 0;
			}
			int var = get_string(module, varname);
			if (var == 0)
			{
				log_print(LOG_ERROR, "Error: Undeclared variable \"%s\".\n", varname);
				free(ctrlstack);
				free(varname);
				return 0;
			}
			bytecode_add_set(module, var);
			
			// Start loop
			bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id);
			
			// Compare with end value
			length = read_word(file, buffer);
			if (!parse_expression_part(module, function, file, 0))
			{
				free(ctrlstack);
				free(varname);
				return 0;
			}
			bytecode_add_get(module, var);
			bytecode_add_cmp(module, CMP_GREATER | CMP_EQUAL);
			
			bytecode_add_if(module, ctrlstack[ctrlstacksize - 1].id + 1);
			
			// Save loop options
			ctrlstack[ctrlstacksize - 1].value1 = var;
			ctrlstack[ctrlstacksize - 1].value2 = 1;
			
			free(varname);
		}
		else if (!strcmp(buffer, "while"))
		{
			// While
			ctrlstack = (ctrl_frame_t*)realloc(ctrlstack, sizeof(ctrl_frame_t) * (ctrlstacksize + 1));
			ctrlstack[ctrlstacksize].id = labelid;
			ctrlstack[ctrlstacksize].type = CTRL_FRAME_WHILE;
			ctrlstacksize++;
			labelid += 2;
			// Read expression
			length = read_word(file, buffer);
			if ((length == -1) || !isalpha(buffer[0]))
			{
				log_print(LOG_ERROR, "Error: Expected variable name after for.\n");
				free(ctrlstack);
				return 0;
			}
			// Start loop
			bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id);
			// Parse expression
			if (!parse_expression_part(module, function, file, 0))
			{
				return 0;
			}
			bytecode_add_if(module, ctrlstack[ctrlstacksize - 1].id + 1);
		}
		else if (!strcmp(buffer, "break"))
		{
			// Break
		}
		else if (!strcmp(buffer, "continue"))
		{
			// Continue
		}
		else if (!strcmp(buffer, "return"))
		{
			// Return
			length = read_word(file, buffer);
			if (length == -1)
			{
				log_print(LOG_ERROR, "Error: Expected expression or ';' after return.\n");
				free(ctrlstack);
				return 0;
			}
			if (strcmp(buffer, ";"))
			{
				// Return value
				
				if (!parse_expression_part(module, function, file, 0))
				{
					return 0;
				}
				if (function->rettype == 0)
				{
					log_print(LOG_WARNING, "Warning: Return with value in function which returns nothing.\n");
				}
				bytecode_add_return(module, 1);
			}
			else
			{
				// No return value
				if (function->rettype != 0)
				{
					log_print(LOG_WARNING, "Warning: Return without value in function with return value.\n");
				}
				bytecode_add_return(module, 0);
			}
		}
		else if (!strcmp(buffer, "end"))
		{
			// End
			log_print(LOG_DEBUG, "End: %d\n", ctrlstacksize);
			if (ctrlstacksize == 0)
			{
				// End of function
				// TODO: Return value
				break;
			}
			
			// End of control structure
			if (ctrlstack[ctrlstacksize - 1].type == CTRL_FRAME_IF)
			{
				bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id);
			}
			else if (ctrlstack[ctrlstacksize - 1].type == CTRL_FRAME_ELSE)
			{
				bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id);
			}
			else if (ctrlstack[ctrlstacksize - 1].type == CTRL_FRAME_FOR)
			{
				// Increment counter
				bytecode_add_get(module, ctrlstack[ctrlstacksize - 1].value1);
				bytecode_add_getc(module, ctrlstack[ctrlstacksize - 1].value2);
				module->bytecode[module->codelen++] = CMD_ADD;
				bytecode_add_set(module, ctrlstack[ctrlstacksize - 1].value1);
				// Jump back
				bytecode_add_jmp(module, ctrlstack[ctrlstacksize - 1].id);
				bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id + 1);
			}
			else if (ctrlstack[ctrlstacksize - 1].type == CTRL_FRAME_WHILE)
			{
				// Jump back
				bytecode_add_jmp(module, ctrlstack[ctrlstacksize - 1].id);
				bytecode_add_label(module, ctrlstack[ctrlstacksize - 1].id + 1);
			}
			
			ctrlstacksize--;
			ctrlstack = (ctrl_frame_t*)realloc(ctrlstack, sizeof(ctrl_frame_t) * ctrlstacksize);
		}
		else
		{
			// Something else
			if (!parse_expression(vm, module, file, function))
			{
				log_print(LOG_ERROR, "Error: Could not parse expression.\n");
				free(ctrlstack);
				return 0;
			}
		}
		
		length = read_word(file, buffer);
		while (length == 0)
		{
			length = read_word(file, buffer);
		}
	}
	free(ctrlstack);
	if (length == -1)
	{
		log_print(LOG_ERROR, "Error: Unexpected EOF.\n");
		return 0;
	}
	
	module->bytecode[module->codelen++] = CMD_END;
	
	return 1;
}
int parse_struct(module_t *module, FILE *file)
{
	// TODO
	return 0;
}

module_t *parse_source(struct ps_vm_t *vm, const char *srcfile)
{
	// Open input file
	FILE *input = fopen(srcfile, "rt");
	if (!input)
	{
		log_print(LOG_ERROR, "Error: Could not open input file \"%s\".\n", srcfile);
		return 0;
	}
	
	module_t *newmod = (module_t*)malloc(sizeof(module_t));
	memset(newmod, 0, sizeof(module_t));
	
	// TODO: Large words crash program
	char buffer[1024];
	// Parse file
	int length;
	while ((length = read_word(input, buffer)) > 0)
	{
		if (length > 0)
		{
			if (!strcmp(buffer, "function"))
			{
				// Function
				log_print(LOG_DEBUG, "Function.\n");
				if (!parse_function(vm, newmod, input))
				{
					log_print(LOG_ERROR, "Could not parse function.\n");
					break;
				}
			}
			else if (!strcmp(buffer, "struct"))
			{
				// Struct
				log_print(LOG_ERROR, "Struct.\n");
				break; // TODO
			}
			else if (isalpha(buffer[0]))
			{
				// Variable type
				char *vartype = strdup(buffer);
				if (((length = read_word(input, buffer)) == -1) || !isalpha(buffer[0]))
				{
					log_print(LOG_ERROR, "Error: Expected variable name, got \"%s\",\"%s\".\n", vartype, buffer);
					free(vartype);
					break;
				}
				// Variable name
				char *varname = strdup(buffer);
				log_print(LOG_DEBUG, "Variable: \"%s %s\"\n", vartype, varname);
				if (((length = read_word(input, buffer)) == -1) || (strcmp(buffer, "=") && strcmp(buffer, ";")))
				{
					log_print(LOG_ERROR, "Error: Expected ';' oder '=' after variable declaration.\n");
					free(vartype);
					free(varname);
					break;
				}
				
				int value = 0;
				char *valuestr = 0;
				
				if (!strcmp(buffer, "="))
				{
					// Value
					if (!strcmp(vartype, "int"))
					{
						if (((length = read_word(input, buffer)) == -1) || !isdigit(buffer[0]))
						{
							log_print(LOG_WARNING, "Warning: Wrong default value type, default value ignored.\n");
						}
						else
						{
							// TODO: other types
							value = atoi(buffer);
							log_print(LOG_DEBUG, "Integer value: %d\n", value);
						}
					}
					else if (!strcmp(vartype, "string"))
					{
						if (((length = read_word(input, buffer)) == -1) || (buffer[0] != '"'))
						{
							log_print(LOG_WARNING, "Warning: Wrong default value type, default value ignored.\n");
						}
						else
						{
							valuestr = strdup(buffer + 1);
							valuestr[strlen(valuestr) - 1] = 0;
						}
					}
					else
					{
						log_print(LOG_WARNING, "Warning: Default value for variable type not supported.\n");
						// Discard value
						length = read_word(input, buffer);
					}
					length = read_word(input, buffer);
					if ((length == -1) || strcmp(buffer, ";"))
					{
						log_print(LOG_ERROR, "Error: Expected ';'.\n");
					}
				}
				
				// Save variable
				newmod->vars = (mod_var_t*)realloc(newmod->vars, sizeof(mod_var_t) * (newmod->varcount + 1));
				newmod->vars[newmod->varcount].type_name = save_string(newmod, vartype);
				newmod->vars[newmod->varcount].name = save_string(newmod, varname);
				newmod->vars[newmod->varcount].type = get_type(vm, newmod, vartype);
				newmod->vars[newmod->varcount].addr = 0;
				if (value)
				{
					newmod->vars[newmod->varcount].value = value;
				}
				else if (valuestr)
				{
					newmod->vars[newmod->varcount].value = save_string(newmod, valuestr);
					free(valuestr);
				}
				newmod->varcount++;
				
				free(vartype);
				free(varname);
			}
			else if (buffer[0] != ';')
			{
				log_print(LOG_ERROR, "Error: Expected keyword.\n");
			}
		}
	}
	
	// Fill initial variable data
	#ifdef __amd64__
	long long *data = (long long*)newmod->data;
	#else
	int *data = (int*)newmod->data;
	#endif
	int i;
	for (i = 0; i < newmod->varcount; i++)
	{
		if (newmod->vars[i].type == VAR_TYPE_STRUCT)
		{
			// Create struct
			struct_info_t *info = vm_get_struct_info(vm, newmod->strings[newmod->vars[i].type_name]);
			if (info)
			{
				// Allocate memory
				newmod->vars[i].addr = malloc(info->size);
				log_print(LOG_DEBUG, "Allocated %p\n", newmod->vars[i].addr);
				memset(newmod->vars[i].addr, 0, info->size);
				// Call constructor
				if (info->constructor)
				{
					info->constructor(newmod->vars[i].addr);
				}
				#ifdef __amd64__
				data[i] = (long long)newmod->vars[i].addr;
				#else
				data[i] = (int)newmod->vars[i].addr;
				#endif
			}
			else
			{
				log_print(LOG_ERROR, "Error: Structure not found (\"%s\").\n", newmod->strings[newmod->vars[i].type_name]);
				data[i] = 0;
			}
		}
		else
		{
			data[i] = newmod->vars[i].value;
		}
	}
	
	return newmod;
}
int save_module(module_t *module, const char *file)
{
	FILE *output = fopen(file, "wt");
	if (!output)
	{
		log_print(LOG_ERROR, "Error: Could not open output file \"%s\".\n", file);
		return 0;
	}
	
	fwrite(&module->codelen, sizeof(int), 1, output);
	fwrite(module->bytecode, 1, module->codelen, output);
	
	fclose(output);
	
	return 1;
}
void delete_module(struct ps_vm_t *vm, module_t *module)
{
	int i;
	// Compiled program
	if (module->program)
	{
		#ifdef __unix__
		munmap(module->program, 0x1000);
		#else
		free(module->program);
		#endif
		
	}
	for (i = 0; i < module->funccount; i++)
	{
		if (module->functions[i].call_stub)
		{
			#ifdef __unix__
			munmap(module->functions[i].call_stub, 0x1000);
			#else
			free(module->functions[i].call_stub);
			#endif
		}
	}
	
	
	// Variables
	for (i = 0; i < module->varcount; i++)
	{
		if ((module->vars[i].type == VAR_TYPE_STRUCT) && module->vars[i].addr)
		{
			// Create struct
			struct_info_t *info = vm_get_struct_info(vm, module->strings[module->vars[i].type_name]);
			if (info)
			{
				// Call destructor
				if (info->destructor)
				{
					info->destructor(module->vars[i].addr);
				}
				// Free memory
				free(module->vars[i].addr);
			}
		}
	}
	free(module->vars);
	
	// Strings
	for (i = 0; i < module->stringcount; i++)
	{
		free(module->strings[i]);
	}
	free(module->strings);
	// Functions
	for (i = 0; i < module->funccount; i++)
	{
		free(module->functions[i].argnames);
		if (module->functions[i].vars)
		{
			free(module->functions[i].vars);
		}
	}
	free(module->functions);
	
	free(module);
}

