
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "parser.h"
#include <ctype.h>
#include <string.h>

int read_word(FILE *file, char *buffer)
{
	int c;
	// Drop whitespaces
	c = fgetc(file);
	while (strchr(" \n\t", c))
	{
		c = fgetc(file);
	}
	if (c == EOF) return -1;
	
	if (c == '"')
	{
		// String
		int length = 1;
		buffer[0] = c;
		c = fgetc(file);
		while (c != '"')
		{
			if (c == EOF)
			{
				printf("Error: Could not read string.\n");
				return -1;
			}
			if (c == '\\')
			{
				// Escaped character
				c = fgetc(file);
				if (c == EOF)
				{
					printf("Error: Could not read string.\n");
					return -1;
				}
				switch (c)
				{
					case 'n':
						buffer[length] = '\n';
						break;
					case 'r':
						buffer[length] = '\r';
						break;
					case 't':
						buffer[length] = '\t';
						break;
					case '\"':
						buffer[length] = '\"';
						break;
					default:
						printf("Warning: Unknown escape sequence %%%c.\n", c);
						buffer[length] = c;
						break;
				}
				length++;
			}
			else
			{
				// Normal character
				buffer[length] = c;
				length++;
			}
			c = fgetc(file);
		}
		buffer[length++] = '\"';
		buffer[length] = 0;
		return length;
	}
	else if (strchr("()[]", c))
	{
		// Single character
		buffer[0] = c;
		buffer[1] = 0;
		return 1;
	}
	else if (c == '/')
	{
		// Comment or /, /=
		c = fgetc(file);
		if (c == EOF)
		{
			printf("Error: Unexpected EOF.\n");
			return -1;
		}
		if (c == '*')
		{
			// Multiline comment
			int prevc = fgetc(file);
			if (prevc == EOF)
			{
				printf("Error: Unexpected EOF.\n");
				return -1;
			}
			c = fgetc(file);
			while ((c != EOF) && ((c != '/') || (prevc != '*')))
			{
				prevc = c;
				c = fgetc(file);
			}
			if (c == EOF)
			{
				printf("Error: Unexpected EOF.\n");
				return -1;
			}
			return 0;
		}
		else if (c == '/')
		{
			// Single line comment
			c = fgetc(file);
			while ((c != '\n') && (c != EOF))
			{
				c = fgetc(file);
			}
			return 0;
		}
		else if (c == '=')
		{
			// /=
			buffer[0] = '/';
			buffer[1] = '=';
			buffer[2] = 0;
			return 2;
		}
		else
		{
			// /
			buffer[0] = '/';
			buffer[1] = 0;
			fseek(file, -1, SEEK_CUR);
			return 1;
		}
	}
	else if (strchr("+-*=!<>", c))
	{
		// Operator
		int op = c;
		c = fgetc(file);
		if (c == EOF)
		{
			printf("Error: Unexpected EOF.\n");
			return -1;
		}
		if (c == '=')
		{
			buffer[0] = op;
			buffer[1] = '=';
			buffer[2] = 0;
			return 2;
		}
		else
		{
			buffer[0] = op;
			buffer[1] = 0;
			fseek(file, -1, SEEK_CUR);
			return 1;
		}
	}
	else if (isalpha(c) || (c == '_'))
	{
		// Normal word
		int length = 1;
		buffer[0] = c;
		c = fgetc(file);
		while (isalnum(c) || (c == '_'))
		{
			buffer[length] = c;
			length++;
			c = fgetc(file);
		}
		fseek(file, -1, SEEK_CUR);
		buffer[length] = 0;
		return length;
	}
	else if (isdigit(c))
	{
		// Digit
		int length = 1;
		buffer[0] = c;
		c = fgetc(file);
		while (isdigit(c) || (c == '.'))
		{
			buffer[length] = c;
			length++;
			c = fgetc(file);
		}
		fseek(file, -1, SEEK_CUR);
		buffer[length] = 0;
		return length;
	}
	else if (strchr(",;:", c))
	{
		buffer[0] = c;
		buffer[1] = 0;
		return 1;
	}
	else if (c == '.')
	{
		buffer[0] = '.';
		c = fgetc(file);
		if (c == '.')
		{
			buffer[1] = '.';
			buffer[2] = 0;
			return 2;
		}
		fseek(file, -1, SEEK_CUR);
		buffer[1] = 0;
		return 1;
	}
	else
	{
		printf("Error: Invalid token.\n");
		return -1;
	}
}

