
/*
Copyright (c) 2008 Mathias Gottschlag

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifdef __i386__
#include "x86.h"
#include "bytecode.h"
#include "logger.h"

#include <stdio.h>

void x86_set_local_var(module_t *module, int var, int reg)
{
	module->program[module->proglen++] = 0x89; // mov %eax, -var * 4(%ebp)
	switch (reg)
	{
		case REG_EAX:
			module->program[module->proglen++] = 0x45;
			break;
		case REG_EDX:
			module->program[module->proglen++] = 0x55;
			break;
		case REG_ESI:
			module->program[module->proglen++] = 0x75;
			break;
		case REG_EDI:
			module->program[module->proglen++] = 0x7d;
			break;
		default:
			log_print(LOG_ERROR, "Error: x86_set_local_var: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x100 - ((var + 1) * 4);
}
void x86_get_local_var(module_t *module, int var, int reg)
{
	module->program[module->proglen++] = 0x8b; // mov -var * 4(%ebp), %eax
	switch (reg)
	{
		case REG_EAX:
			module->program[module->proglen++] = 0x45;
			break;
		case REG_EDX:
			module->program[module->proglen++] = 0x55;
			break;
		case REG_ESI:
			module->program[module->proglen++] = 0x75;
			break;
		case REG_EDI:
			module->program[module->proglen++] = 0x7d;
			break;
		default:
			log_print(LOG_ERROR, "Error: x86_get_local_var: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x100 - ((var + 1) * 4);
}
void x86_get_offset(module_t *module, int addrreg, int offset, int destreg, int size)
{
	if ((addrreg != REG_EAX) || (destreg != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: amd64_get_offset: Register not implemented.\n");
	}
	if (size < 8)
	{
		module->program[module->proglen++] = 0xb9; // mov $0, %ecx
		*((int*)&module->program[module->proglen + 4]) = 0;
		module->proglen += sizeof(int);
	}
	switch (size)
	{
		case 1:
			module->program[module->proglen++] = 0x8a; // mov offset(%eax), %cl
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 2:
			module->program[module->proglen++] = 0x66; // mov offset(%eax), %cx
			module->program[module->proglen++] = 0x8b;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 4:
			module->program[module->proglen++] = 0x8b; // mov offset(%eax), %ecx
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 8:
			log_print(LOG_ERROR, "Error: Member size 8 not supported on x86!\n");
			break;
	}
}
void x86_set_offset(module_t *module, int addrreg, int offset, int valuereg, int size)
{
	switch (size)
	{
		case 1:
			module->program[module->proglen++] = 0x88; // mov offset(%eax), %cl
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 2:
			module->program[module->proglen++] = 0x66; // mov offset(%eax), %cx
			module->program[module->proglen++] = 0x89;
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 4:
			module->program[module->proglen++] = 0x89; // mov offset(%eax), %ecx
			module->program[module->proglen++] = 0x48;
			module->program[module->proglen++] = offset;
			break;
		case 8:
			printf("Error: Member size 8 not supported on x86!\n");
			break;
	}
}
void x86_call_function(module_t *module, void *function)
{
	module->program[module->proglen++] = 0xb8; // mov $function, %eax
	*((void**)&module->program[module->proglen]) = function;
	module->proglen += sizeof(void*);
	module->program[module->proglen++] = 0xff; // call *%eax
	module->program[module->proglen++] = 0xd0;
}
void x86_add(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %eax
		module->program[module->proglen++] = 0x59; // pop %ecx
		module->program[module->proglen++] = 0x01; // add %ecx, %eax
		module->program[module->proglen++] = 0xc8;
		module->program[module->proglen++] = 0x50; // push %eax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		x86_adjust_esp(module, 4);
		module->program[module->proglen++] = 0xd8; // fadd (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: x86_add: Type not implemented.\n");
	}
}
void x86_sub(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %eax
		module->program[module->proglen++] = 0x59; // pop %ecx
		module->program[module->proglen++] = 0x29; // sub %ecx, %eax
		module->program[module->proglen++] = 0xc8;
		module->program[module->proglen++] = 0x50; // push %eax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		x86_adjust_esp(module, 4);
		module->program[module->proglen++] = 0xd8; // fsub (%esp)
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: x86_sub: Type not implemented.\n");
	}
}
void x86_mul(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %eax
		module->program[module->proglen++] = 0x59; // pop %ecx
		module->program[module->proglen++] = 0xf7; // imul %ecx
		module->program[module->proglen++] = 0xe8;
		module->program[module->proglen++] = 0x50; // push %eax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		x86_adjust_esp(module, 4);
		module->program[module->proglen++] = 0xd8; // fmul (%esp)
		module->program[module->proglen++] = 0x0c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: x86_mul: Type not implemented.\n");
	}
}
void x86_div(module_t *module, int reg1, int reg2, int type)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_add: Register not implemented.\n");
	}
	if (type == VAR_TYPE_INT)
	{
		module->program[module->proglen++] = 0x58; // pop %eax
		module->program[module->proglen++] = 0x59; // pop %ecx
		module->program[module->proglen++] = 0xba; // mov $0, %edx
		*((int*)&module->program[module->proglen]) = 0;
		module->proglen += sizeof(int);
		module->program[module->proglen++] = 0xf7; // idiv %ecx
		module->program[module->proglen++] = 0xf8;
		module->program[module->proglen++] = 0x50; // push %eax
	}
	else if (type == VAR_TYPE_FLOAT)
	{
		module->program[module->proglen++] = 0xd9; // flds (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		x86_adjust_esp(module, 4);
		module->program[module->proglen++] = 0xd8; // fdiv (%esp)
		module->program[module->proglen++] = 0x34;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
	}
	else
	{
		log_print(LOG_ERROR, "Error: x86_div: Type not implemented.\n");
	}
}
void x86_and(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_and: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %eax
	module->program[module->proglen++] = 0x59; // pop %ecx
	module->program[module->proglen++] = 0x21; // and %ecx, %eax
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %eax
}
void x86_or(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_or: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %eax
	module->program[module->proglen++] = 0x59; // pop %ecx
	module->program[module->proglen++] = 0x09; // and %ecx, %eax
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %eax
}
void x86_xor(module_t *module, int reg1, int reg2)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_xor: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %eax
	module->program[module->proglen++] = 0x59; // pop %ecx
	module->program[module->proglen++] = 0x31; // and %ecx, %eax
	module->program[module->proglen++] = 0xc8;
	module->program[module->proglen++] = 0x50; // push %eax
}
void x86_compare(module_t *module, int reg1, int reg2, int type, int vartype)
{
	if ((reg1 != REG_EAX) || (reg2 != REG_ECX))
	{
		log_print(LOG_ERROR, "Error: x86_compare: Register not implemented.\n");
	}
	module->program[module->proglen++] = 0x58; // pop %eax
	module->program[module->proglen++] = 0x59; // pop %ecx
	module->program[module->proglen++] = 0x6a; // push $0
	module->program[module->proglen++] = 0x00;
	module->program[module->proglen++] = 0x39; // cmp %eax, %ecx
	module->program[module->proglen++] = 0xc1;
	
	if (type == CMP_SMALLER)
	{
		module->program[module->proglen++] = 0x7d; // jge
	}
	else if (type == CMP_GREATER)
	{
		module->program[module->proglen++] = 0x7e; // jle
	}
	else if (type == (CMP_SMALLER | CMP_EQUAL))
	{
		module->program[module->proglen++] = 0x7f; // jg
	}
	else if (type == (CMP_GREATER | CMP_EQUAL))
	{
		module->program[module->proglen++] = 0x7c; // jl
	}
	else if (type == (CMP_GREATER | CMP_SMALLER))
	{
		module->program[module->proglen++] = 0x74; // je
	}
	else if (type == CMP_EQUAL)
	{
		module->program[module->proglen++] = 0x75; // jne
	}
	module->program[module->proglen++] = 0x04;
	
	module->program[module->proglen++] = 0xc6; // mov $1, (%esp)
	module->program[module->proglen++] = 0x04;
	module->program[module->proglen++] = 0x24;
	module->program[module->proglen++] = 0x01;
}

void x86_adjust_esp(module_t *module, int size)
{
	if (size > 0xff)
	{
		module->program[module->proglen++] = 0x81; // add $size, %esp
		module->program[module->proglen++] = 0xc4;
		*((int*)&module->program[module->proglen]) = size;
		module->proglen += sizeof(int);
	}
	else if (size > 0)
	{
		module->program[module->proglen++] = 0x83; // add $size, %esp
		module->program[module->proglen++] = 0xc4;
		module->program[module->proglen++] = size;
	}
	else if (size < 0)
	{
		module->program[module->proglen++] = 0x81; // sub $size, %esp
		module->program[module->proglen++] = 0xec;
		*((int*)&module->program[module->proglen]) = -size;
		module->proglen += sizeof(int);
	}
}

void x86_convert(module_t *module, int srctype, int desttype)
{
	log_print(LOG_DEBUG, "Converting %d to %d.\n", srctype, desttype);
	if ((srctype == VAR_TYPE_FLOAT) && (desttype == VAR_TYPE_INT))
	{
		module->program[module->proglen++] = 0x50; // push %eax
		module->program[module->proglen++] = 0xd9; // flds (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xdb; // fistp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x58; // pop %eax
	}
	else if ((srctype == VAR_TYPE_INT) && (desttype == VAR_TYPE_FLOAT))
	{
		module->program[module->proglen++] = 0x50; // push %eax
		module->program[module->proglen++] = 0xdf; // fild (%esp)
		module->program[module->proglen++] = 0x04;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0xd9; // fstp (%esp)
		module->program[module->proglen++] = 0x1c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = 0x58; // pop %eax
	}
	else
	{
		log_print(LOG_WARNING, "Warning: x86_convert: Cannot convert %d to %d.\n", srctype, desttype);
	}
}

void x86_flip_args(module_t *module, int count, int skip, int convert, int *types, int *expected)
{
	int i;
	for (i = 0; i < count / 2; i++)
	{
		// TODO: Optimize this
		// Swap first and last argument
		module->program[module->proglen++] = 0x8b; // mov i(%esp),%eax
		module->program[module->proglen++] = 0x44;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = (i + skip) * 4;
		if (convert)
		{
			if (types[count - 1 - i] != expected[count - 1 - i])
			{
				x86_convert(module, types[count - 1 - i], expected[count - 1 - i]);
			}
		}
		module->program[module->proglen++] = 0x89; // mov %eax, %ecx
		module->program[module->proglen++] = 0xc1;
		module->program[module->proglen++] = 0x8b; // mov argcount-i(%esp),%eax
		module->program[module->proglen++] = 0x44;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = (count - 1 - i + skip) * 4;
		if (convert)
		{
			if (types[i] != expected[i])
			{
				x86_convert(module, types[i], expected[i]);
			}
		}
		module->program[module->proglen++] = 0x89; // mov %eax, i(%esp)
		module->program[module->proglen++] = 0x44;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = (i + skip) * 4;
		module->program[module->proglen++] = 0x89; // mov %ecx, argcount-i(%esp)
		module->program[module->proglen++] = 0x4c;
		module->program[module->proglen++] = 0x24;
		module->program[module->proglen++] = (count - 1 - i + skip) * 4;
	}
	if (convert && (count % 2))
	{
		if (types[count / 2] != expected[count / 2])
		{
			module->program[module->proglen++] = 0x8b; // mov i(%esp),%eax
			module->program[module->proglen++] = 0x44;
			module->program[module->proglen++] = 0x24;
			module->program[module->proglen++] = (count / 2 + skip) * 4;
			x86_convert(module, types[count / 2], expected[count / 2]);
			module->program[module->proglen++] = 0x89; // mov %eax, i(%esp)
			module->program[module->proglen++] = 0x44;
			module->program[module->proglen++] = 0x24;
			module->program[module->proglen++] = (count / 2 + skip) * 4;
		}
	}
}

void x86_call_operator(module_t *module, int varidx, void *function, int type, char *name, int local)
{
	module->program[module->proglen++] = 0x50; // push %eax
	module->program[module->proglen++] = 0x68; // push $name
	*((char**)&module->program[module->proglen]) = name;
	module->proglen += sizeof(char*);
	module->program[module->proglen++] = 0x68; // push $type
	*((int*)&module->program[module->proglen]) = type;
	module->proglen += sizeof(int);
	if (local)
	{
		module->program[module->proglen++] = 0xff;
		module->program[module->proglen++] = 0x75;
		module->program[module->proglen++] = 0x100 - ((varidx + 1) * 4);
	}
	else
	{
		module->program[module->proglen++] = 0xff; // push var
		module->program[module->proglen++] = 0x35;
		void *addr = module->data + varidx * 4;
		*((void**)&module->program[module->proglen]) = addr;
		module->proglen += sizeof(void*);
	}
	x86_call_function(module, function);
	x86_adjust_esp(module, 16);
}

#endif

